#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **warning** (i.e., non-fatal errors with associated types emitted by
the standard :mod:`warnings` module) facilities.
'''

# ....................{ IMPORTS                            }....................
import warnings
from betse.util.type.types import type_check, ClassType, GeneratorType
from contextlib import contextmanager

# ....................{ MANAGERS                           }....................
@contextmanager
@type_check
def ignoring_warnings(warning_class: ClassType = Warning) -> GeneratorType:
    '''
    Single-shot context manager temporarily ignoring all warnings of the passed
    type emitted by the :mod:`warnings` module for the duration of this context.

    Caveats
    -----------
    This context manager is single-shot and hence *not* safely reusable across
    multiple ``with`` blocks. Why? Because the low-level
    :class:`warnings.catch_warnings` class to which this context manager defers
    is itself single-shot, explicitly raising exceptions on reuse. While
    lamentable, this overhead appears to be unavoidable.

    Parameters
    -----------
    warning_class : optional[ClassType]
        Base class of all warnings to be ignored by this context manager.
        Defaults to :class:`Warning`, the superclass of all warning
        subclasses (builtin and user-defined), in which case this context
        manager will ignore *all* warnings.
    '''

    # Isolate all side effects produced by the "warnings" module to this block.
    with warnings.catch_warnings():
        # Ignore all warnings emitted by this module that are instances of the
        # previously passed base class.
        warnings.simplefilter(action='ignore', category=warning_class)

        # Yield control to the body of the caller's "with" block.
        yield



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Apple macOS-specific facilities.
'''

# ....................{ IMPORTS                            }....................
from betse.exceptions import BetseOSException
from betse.util.io.log import logs
from ctypes import CDLL, byref, c_int

# ....................{ CONSTANTS                          }....................
_SECURITY_FRAMEWORK_DYLIB_FILENAME = (
    '/System/Library/Frameworks/Security.framework/Security')
'''
Absolute path of the system-wide `Security.framework` Macho-O shared library
providing the macOS-specific security context for the current process.

This library is dynamically loadable into the address space of the current
process with the :class:`ctypes.CDLL` class. Since all Macho-O shared libraries
necessarily have the filetype `dylib`, this filetype is safely omitted here.
'''


_SECURITY_SESSION_ID_CURRENT = -1
'''
Magic integer defined as `callerSecuritySession` by the macOS-specific
`/System/Library/Frameworks/Security.Framework/Headers/AuthSession.h` C header
suitable for passing to C functions accepting parameters of C type
`SecuritySessionId` (e.g., `SessionGetInfo()`).

When passed, this integer signifies the **current security session** (i.e., the
the security session to which the current process belongs).

See Also
----------
https://opensource.apple.com/source/libsecurity_authorization/libsecurity_authorization-32564/lib/AuthSession.h
    C header defining this magic integer.
'''


_SECURITY_SESSION_HAS_GRAPHIC_ACCESS = 0x0010
'''
Bit flag defined as `sessionHasGraphicAccess` by the macOS-specific
`/System/Library/Frameworks/Security.Framework/Headers/AuthSession.h` C header
masking the attributes bit field returned by the `SessionGetInfo()` C function
also declared by that header.

When enabled, this bit signifies the current process to have access to the Aqua
display server and hence be headfull (rather than headless).

See Also
----------
https://opensource.apple.com/source/libsecurity_authorization/libsecurity_authorization-32564/lib/AuthSession.h
    C header defining this bit flag.
'''

# ....................{ EXCEPTIONS                         }....................
def die_unless_macos() -> None:
    '''
    Raise an exception unless the current platform is Apple macOS.
    '''

    # Avoid circular import dependencies.
    from betse.util.os import oses

    # If the current platform is *NOT* macOS, raise an exception.
    if not oses.is_macos():
        raise BetseOSException(
            'Current platform {} not macOS.'.format(oses.get_name()))

# ....................{ TESTERS                            }....................
def is_aqua() -> bool:
    '''
    `True` only if the current process has access to the Aqua display server
    specific to macOS, implying this process to be headfull and hence support
    both CLIs and GUIs.

    See Also
    ----------
    https://developer.apple.com/library/content/technotes/tn2083/_index.html#//apple_ref/doc/uid/DTS10003794-CH1-SUBSECTION19
        "Security Context" subsection of "Technical Note TN2083: Daemons and
        Agents," a psuedo-human-readable discussion of the
        `sessionHasGraphicAccess` bit flag returned by the low-level
        `SessionGetInfo()` C function.
    '''

    # Avoid circular import dependencies.
    from betse.util.path import files
    from betse.util.path.command.cmdexit import SUCCESS

    # Raise an exception unless the current platform is macOS.
    die_unless_macos()

    # Attempt all of the following in a safe manner catching, logging, and
    # converting exceptions into a false return value. This tester is *NOT*
    # mission-critical and hence should *NOT* halt the application on
    # library-specific failures.
    try:
        # If the system-wide Macho-O shared library providing the macOS
        # security context for the current process does *NOT* exist (after
        # following symbolic links), raise an exception.
        files.die_unless_file(_SECURITY_FRAMEWORK_DYLIB_FILENAME)

        # Dynamically load this library into the address space of this process.
        security_framework = CDLL(_SECURITY_FRAMEWORK_DYLIB_FILENAME)

        # Possibly non-unique identifier of the security session to request the
        # attributes of, signifying that of the current process.
        session_id = _SECURITY_SESSION_ID_CURRENT

        # Unique identifier of the requested security session, returned
        # by reference from the SessionGetInfo() C function called below. This
        # identifier is useless for our purposes and hence ignored below.
        session_id_real = c_int(0)

        # Attributes bit field of the requested security session, returned by
        # reference from the SessionGetInfo() C function called below.
        session_attributes = c_int(0)

        # C-style error integer returned by calling the SessionGetInfo() C
        # function exported by this Macho-O shared library, passing:
        #
        # * The input non-unique session identifier by value.
        # * The output unique session identifier by reference.
        # * The output session attributes integer by reference.
        session_errno = security_framework.SessionGetInfo(
            session_id, byref(session_id_real), byref(session_attributes))

        # This process has access to the Aqua display server if and only if...
        return (
            # The above function call succeeded *AND*...
            session_errno == SUCCESS and
            # The session attributes bit field returned by this call has the
            # corresponding bit flag enabled.
            session_attributes.value & _SECURITY_SESSION_HAS_GRAPHIC_ACCESS
        )

    # If the above logic fails with any exception...
    except Exception as exc:
        # Log a non-fatal warning informing users of this failure.
        logs.log_warning(
            'macOS-specific SessionGetInfo() C function failed: {}'.format(
                exc.strerror))

        # Assume this process to *NOT* have access to the Aqua display server.
        return False



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Fixtures initializing BETSE in a manner suitable for _all_ testing.
'''

# ....................{ IMPORTS                            }....................
from pytest import fixture

# ....................{ FIXTURES                           }....................
@fixture(scope='session', autouse=True)
def betse_init() -> None:
    '''
    Automatically run per-session fixture initializing BETSE in a manner
    suitable for unit testing.

    Specifically, this fixture:

    * Initializes the BETSE core.
    * Initializes all third-party dependencies thereof.
    * Increases logging verbosity to the maximum level (i.e., ``DEBUG``).
    * Enables the default non-interactive matplotlib backend ``Agg``,
      *guaranteed* to be usable on all platforms. By default, matplotlib enables
      an interactive backend (e.g., ``Qt5Agg``) unsuitable for use under
      typically headless test automation.

    Motivation
    ----------
    This fixture is automatically requested by *all* tests (functional, unit,
    or otherwise) without needing to be explicitly requested. Moreover, this
    fixture is imported into the top-level :mod:`betse_test.conftest` plugin and
    hence *guaranteed* to be run prior to all other BETSE-specific fixtures.
    Doing so avoids spurious issues in other fixtures and tests.

    Notably, the early test-specific initialization of both BETSE and
    dependencies prevents the magic :func:`betse.science._ignite` function from
    attempting to perform a subsequent test-agnostic initialization of either
    BETSE or dependencies on the first importation of the :mod:`betse.science`
    subpackage, as occurs in fixtures importing from that subpackage (e.g., the
    :mod:`betse_test.fixture.simconf.simconfer` fixture importing the
    :mod:`betse_test.fixture.simconf.simconfwrapper` submodule importing the
    :med:`betse.science.config.confwrap` submodule).
    '''

    # Defer heavyweight imports.
    from betse import ignition
    from betse.lib import libs
    from betse.util.io.log import logconfig

    # Inform users of this initialization.
    print('\n[py.test] Initializing BETSE...')

    # Initialize the core application. Note that the higher-level
    # ignition.ignote() function is intentionally *NOT* called here, as doing so
    # could erroneously enable a headfull matplotlib backend.
    ignition.init()

    # Singleton logging configuration for the current Python process retrieved
    # *AFTER* initializing the core application.
    log_config = logconfig.get()

    # Increase logging verbosity to the maximum level (i.e., "DEBUG").
    log_config.is_verbose = True

    # Initialize all dependencies *AFTER* both the core application and logging.
    libs.init(matplotlib_backend_name='Agg')



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
CLI-specific functional tests exporting simulation configurations to external
directories, typically as a subprocess of a larger functional test (e.g., to
validate application backward compatibility).
'''

# ....................{ IMPORTS                            }....................
import pytest
from betse_test.util import requests
from betse_test.util.mark.skip import skip_if

# ....................{ OPTIONS                            }....................
EXPORT_SIM_CONF_DIRNAME = pytest.config.option.export_sim_conf_dirname
'''
Absolute or relative path of the target directory to export (i.e., recursively
copy) each source simulation configuration directory into if any *or* ``None``
otherwise.

See Also
----------
:func:`betse_test.conftest.pytest_addoption`
    Further details.
'''

# ....................{ DECORATORS                         }....................
skip_unless_export_sim_conf_dir = skip_if(
    EXPORT_SIM_CONF_DIRNAME is None,
    reason='CLI option "--export-sim-conf-dir" unpassed.')
'''
Decorator skipping the decorated test if the ``--export-sim-conf-dir``
command-line option specific to this test suite was *not* passed to the
``py.test`` command.
'''

# ....................{ TESTS                              }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# CAUTION: The name of this test is assumed to *NOT* change across Git commits.
# Changing this name would violate this assumption and hence forwards
# compatibility with future versions of this test suite.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

@skip_unless_export_sim_conf_dir
def test_cli_sim_export(
    request: '_pytest.python.FixtureRequest',
    betse_cli_sim: 'CLISimTester',
) -> None:
    '''
    Functional test exporting the simulation configuration directory produced by
    running all BETSE subcommands exporting pickled objects (i.e., ``seed``,
    ``init``, and ``sim``) to the ``test_sim_export`` subdirectory of the
    ``--export-sim-conf-dir`` command-line option passed to the ``py.test``
    command.

    Usage
    ----------
    This pseudo-test is typically run explicitly from the CLI as follows:

    .. code:: bash

        $ ./test -k test_cli_sim_export --export-sim-conf-dir=~/some/where

    Parameters
    ----------
    request : _pytest.python.FixtureRequest
        Builtin fixture describing this test.
    betse_cli_sim : CLISimTester
        Object running BETSE CLI simulation subcommands.
    '''

    # Defer heavyweight imports.
    from betse.util.path import dirs, pathnames

    # Name of the current test.
    test_name = requests.get_tested_name(request)

    # Absolute or relative pathname of the target directory to export into,
    # canonicalized for caller convenience (e.g., to ensure tilde expansion) and
    # created if this directory does *NOT* already exist.
    trg_sim_conf_parent_dirname = dirs.canonicalize_and_make_unless_dir(
        EXPORT_SIM_CONF_DIRNAME)

    # Absolute or relative pathname of the test-specific target directory to
    # copy the test-specific source simulation configuration directory to.
    trg_sim_conf_dirname = pathnames.join(
        trg_sim_conf_parent_dirname, test_name)

    # If this test-specific target directory already exists, raise an exception.
    dirs.die_if_dir(trg_sim_conf_dirname)

    # Absolute pathname of the test-specific source simulation configuration
    # directory to copy from.
    src_sim_conf_dirname = betse_cli_sim.sim_state.conf_dirname

    # Test all subcommands exporting pickled objects. Note that plotting
    # subcommands export no such objects and hence are intentionally omitted.
    betse_cli_sim.run_subcommands_sim()

    # Export (i.e., recursively copy) this source to target directory.
    dirs.copy(
        src_dirname=src_sim_conf_dirname,
        trg_dirname=trg_sim_conf_dirname)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Unit tests exercising the `betse.util.type.types.type_check` decorator,
implementing a rudimentary subset of PEP 484-style type checking based on Python
3.x function annotations.
'''

# ....................{ IMPORTS                            }....................
import pytest
from random import Random

# ....................{ TESTS                              }....................
def test_type_check_noop() -> None:
    '''
    Test type checking for a function with no function annotations, reducing to
    *no* type checking.
    '''

    # Import this decorator.
    from betse.util.type.types import type_check

    # Unannotated function to be type checked.
    @type_check
    def khorne(gork, mork):
        return gork + mork

    # Call this function and assert the expected return value.
    assert khorne('WAAAGH!', '!HGAAAW') == 'WAAAGH!!HGAAAW'

# ....................{ TESTS ~ pass : param               }....................
def test_type_check_pass_param_positional_and_keyword() -> None:
    '''
    Test type checking for a function call successfully passed annotated
    non-variadic positional and keyword parameters.
    '''

    # Import this decorator.
    from betse.util.type.types import type_check

    # Function to be type checked.
    @type_check
    def slaanesh(daemonette: str, keeper_of_secrets: str) -> str:
        return daemonette + keeper_of_secrets

    # Call this function with both positional and keyword arguments and assert
    # the expected return value.
    assert slaanesh(
        'Seeker of Decadence', keeper_of_secrets="N'Kari") == (
        "Seeker of DecadenceN'Kari")


def test_type_check_pass_param_keyword_only() -> None:
    '''
    Test type checking for a function call successfully passed an annotated
    keyword-only parameter following a variadic positional parameter.
    '''

    # Import this decorator.
    from betse.util.type.types import type_check

    # Function to be type checked.
    @type_check
    def changer_of_ways(
        sky_shark: str, *dark_chronology: int, chaos_spawn: str) -> str:
        return (
            sky_shark +
            str(dark_chronology[0]) +
            str(dark_chronology[-1]) +
            chaos_spawn
        )

    # Call this function and assert the expected return value.
    assert changer_of_ways(
        'Screamers', 0, 1, 15, 25, chaos_spawn="Mith'an'driarkh") == (
        "Screamers025Mith'an'driarkh")


def test_type_check_pass_param_tuple() -> None:
    '''
    Test type checking for a function call successfully passed a parameter
    annotated as a tuple.
    '''

    # Import this decorator.
    from betse.util.type.types import type_check

    # Function to be type checked. For completeness, test both an actual class
    # *AND* a the fully-qualified name of a class in this tuple annotation.
    @type_check
    def genestealer(tyranid: str, hive_fleet: (str, 'int')) -> str:
        return tyranid + str(hive_fleet)

    # Call this function with each of the two types listed in the above tuple.
    assert genestealer('Norn-Queen', 'Behemoth') == 'Norn-QueenBehemoth'
    assert genestealer('Carnifex', 0xDEADBEEF) == 'Carnifex3735928559'


def test_type_check_pass_param_custom() -> None:
    '''
    Test type checking for a function call successfully passed a parameter
    annotated as a user-defined rather than built-in type.
    '''

    # Import this decorator.
    from betse.util.type.types import type_check

    # User-defined type.
    class CustomTestStr(str):
        pass

    # Function to be type checked.
    @type_check
    def hrud(gugann: str, delphic_plague: CustomTestStr) -> str:
        return gugann + delphic_plague

    # Call this function with each of the above type.
    assert hrud(
        'Troglydium hruddi', delphic_plague=CustomTestStr('Delphic Sink')) == (
        'Troglydium hruddiDelphic Sink')


def test_type_check_pass_param_str() -> None:
    '''
    Test type checking for a function call successfully passed a parameter
    annotated as a string.
    '''

    # Import this decorator.
    from betse.util.type.types import type_check

    # Dates between which the Sisters of Battle must have been established.
    ESTABLISHMENT_DATE_MIN = 36000
    ESTABLISHMENT_DATE_MAX = 37000

    # Function to be type checked.
    @type_check
    def sisters_of_battle(
        leader: str, establishment: 'random.Random') -> int:
        return establishment.randint(
            ESTABLISHMENT_DATE_MIN, ESTABLISHMENT_DATE_MAX)

    # Call this function with an instance of the type named above.
    assert sisters_of_battle('Abbess Sanctorum', Random()) in range(
        ESTABLISHMENT_DATE_MIN, ESTABLISHMENT_DATE_MAX + 1)

# ....................{ TESTS ~ pass : return              }....................
def test_type_check_pass_return_none() -> None:
    '''
    Test type checking for a function call successfully returning ``None`` and
    annotated as such.
    '''

    # Import this decorator.
    from betse.util.type.types import type_check

    # Function to be type checked.
    @type_check
    def xenos(interex: str, diasporex: str) -> None:
        print(interex + diasporex)

    # Call this function and assert no value to be returned.
    assert xenos(
        'Luna Wolves', diasporex='Iron Hands Legion') is None

# ....................{ TESTS ~ fail                       }....................
def test_type_check_fail_keyword_unknown() -> None:
    '''
    Test type checking for an annotated function call passed an unrecognized
    keyword parameter.
    '''

    # Import this decorator.
    from betse.util.type.types import type_check

    # Annotated function to be type checked.
    @type_check
    def tau(kroot: str, vespid: str) -> str:
        return kroot + vespid

    # Call this function with an unrecognized keyword parameter and assert the
    # expected exception.
    with pytest.raises(TypeError) as exception:
        tau(kroot='Greater Good', nicassar='Dhow')

    # For readability, this should be a "TypeError" synopsizing the exact issue
    # raised by the Python interpreter on calling the original function rather
    # than a "TypeError" failing to synopsize the exact issue raised by the
    # wrapper type-checking the original function. Since the function
    # annotations defined above guarantee that the exception message of the
    # latter will be suffixed by "not a str", ensure this is *NOT* the case.
    assert not str(exception.value).endswith('not a str')


def test_type_check_fail_param_name() -> None:
    '''
    Test type checking for a function accepting a parameter name reserved for
    use by the `@type_check` decorator.
    '''

    # Import this decorator.
    from betse.util.type.types import type_check

    # Define a function accepting a reserved parameter name and assert the
    # expected exception.
    with pytest.raises(NameError):
        @type_check
        def jokaero(weaponsmith: str, __beartype_func: str) -> str:
            return weaponsmith + __beartype_func

# ....................{ TESTS ~ fail : type                }....................
def test_type_check_fail_param_str() -> None:
    '''
    Test type checking for an annotated function call failing a string
    parameter type check.
    '''

    # Import this decorator.
    from betse.util.type.types import type_check

    # Dates between which the Black Legion must have been established.
    ESTABLISHMENT_DATE_MIN = 30000
    ESTABLISHMENT_DATE_MAX = 31000

    # Function to be type checked.
    @type_check
    def black_legion(primarch: str, establishment: 'random.Random') -> int:
        return establishment.randint(
            ESTABLISHMENT_DATE_MIN, ESTABLISHMENT_DATE_MAX)

    # Call this function with an invalid type and assert the expected exception.
    with pytest.raises(TypeError):
        black_legion('Horus', 'Abaddon the Despoiler')


def test_type_check_fail_param_nonvariadic_type() -> None:
    '''
    Test type checking for an annotated function call failing a non-variadic
    parameter type check.
    '''

    # Import this decorator.
    from betse.util.type.types import type_check

    # Annotated function to be type checked.
    @type_check
    def eldar(isha: str, asuryan: (str, int)) -> str:
        return isha + asuryan

    # Call this function with an invalid type and assert the expected exception.
    with pytest.raises(TypeError):
        eldar('Mother of the Eldar', 100.100)


def test_type_check_fail_param_variadic_type() -> None:
    '''
    Test type checking for an annotated function call failing a variadic
    parameter type check.
    '''

    # Import this decorator.
    from betse.util.type.types import type_check

    # Annotated function to be type checked.
    @type_check
    def imperium_of_man(
        space_marines: str, *ceaseless_years: int, primarch: str) -> str:
        return space_marines + str(ceaseless_years[1]) + primarch

    # Call this function with an invalid type and assert the expected exception.
    with pytest.raises(TypeError):
        imperium_of_man(
            'Legiones Astartes', 30, 31, 36, 'M41', primarch='Leman Russ')


def test_type_check_fail_return_type() -> None:
    '''
    Test type checking for an annotated function call failing a return type
    check.
    '''

    # Import this decorator.
    from betse.util.type.types import type_check

    # Annotated function to be type checked.
    @type_check
    def necron(star_god: str, old_one: str) -> str:
        return 60e6

    # Call this function and assert the expected exception.
    with pytest.raises(TypeError):
        necron("C'tan", 'Elder Thing')

# ....................{ TESTS ~ fail : annotation          }....................
def test_type_check_fail_annotation_param_int() -> None:
    '''
    Test type checking for a function with an unsupported integer parameter
    annotation.
    '''

    # Import this decorator.
    from betse.util.type.types import type_check

    # Assert the expected exception from attempting to type check a function
    # with a parameter annotation that is *NOT* a type.
    with pytest.raises(TypeError):
        @type_check
        def nurgle(nurgling: str, great_unclean_one: 0x8BADF00D) -> str:
            return nurgling + str(great_unclean_one)


def test_type_check_fail_annotation_param_str() -> None:
    '''
    Test type checking for a function with unsupported string parameter
    annotations.
    '''

    # Import this decorator.
    from betse.util.type.types import type_check

    # Assert the expected exception from attempting to type check a function
    # with a string parameter annotation referencing an unimportable module.
    with pytest.raises(ImportError):
        @type_check
        def eye_of_terror(
            ocularis_terribus: str,

            # While highly unlikely that a top-level module with this name will
            # ever exist, the possibility cannot be discounted. Since there
            # appears to be no syntactically valid module name prohibited from
            # existing, this is probably the best we can do.
            segmentum_obscurus: '__rand0m__.Warp',
        ) -> str:
            return ocularis_terribus + segmentum_obscurus

        eye_of_terror('Perturabo', 'Crone Worlds')

    # Assert the expected exception from attempting to type check a function
    # with a string parameter annotation referencing a missing attribute of an
    # importable module.
    with pytest.raises(ImportError):
        @type_check
        def navigator(
            astronomicon: str,

            # While highly unlikely that a top-level module attribute with this
            # name will ever exist, the possibility cannot be discounted. Since
            # there appears to be no syntactically valid module attribute name
            # prohibited from existing, this is probably the best we can do.
            navis_nobilite: 'random.__Psych1cL1ght__',
        ) -> str:
            return astronomicon + navis_nobilite

        navigator('Homo navigo', 'Kartr Hollis')


def test_type_check_fail_annotation_return() -> None:
    '''
    Test type checking for a function with an unsupported return annotation.
    '''

    # Import this decorator.
    from betse.util.type.types import type_check

    # Assert the expected exception from attempting to type check a function
    # with a return annotation that is *NOT* a supported type.
    with pytest.raises(TypeError):
        @type_check
        def tzeentch(disc: str, lord_of_change: str) -> ['Player of Games',]:
            return 0xB16B00B5



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Subsidiary entry point of BETSE's command line interface (CLI), preserving
backward compatibility with prior versions.
'''

# ....................{ IMPORTS                            }....................
from betse.__main__ import main
from betse.util.path.command import cmdexit

# ....................{ MAIN                               }....................
# If this module is imported from the command line, run BETSE's CLI; else, noop.
# For POSIX compliance, the exit status returned by this function is propagated
# to the caller as this script's exit status.
if __name__ == '__main__':
    cmdexit.exit(main())



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

#FIXME: Error messages should be prefixed by strings uniquely identifying the
#sources of those messages. Specifically:
#
#* Warnings should be prefixed by "<Warning> {source_module_basename}".
#* Errors should be prefixed by "<Error> {source_module_basename}".
#
#See the following stackoveflow question for details on how to implement this:
#    https://stackoverflow.com/questions/14844970/modifying-logging-message-format-based-on-message-logging-level-in-python3

#FIXME: The following blog post provides useful instructions on deserializing
#logging settings from a YAML-formatted configuration file. Leveraging such
#instructions, we could permit users to setup logging as they see fit (e.g.,
#configuring the logging file and level). This is probably the preferable means
#of doing so, rather than providing a gamut of CLI options:
#
#    http://victorlin.me/posts/2012/08/26/good-logging-practice-in-python

'''
Low-level logging facilities.

Logging Hierarchy
----------
Loggers are hierarchically structured according to their `.`-delimited names.
Since the name of the root logger is _always_ the empty string, this logger is
_always_ the parent of all user-defined loggers. This hierarchy is an implicit
consequence of logger names and hence requires no manual intervention (e.g., the
root logger `` is implicitly the parent of a user-defined logger `A` is
implicitly the parent of a user-defined logger `A.B`).

By default, logger messages are implicitly propagated up the logger hierarchy
(e.g., messages to logger `A.B` are implicitly progagated to logger `A` are
are implicitly progagated to the root logger). This is a good thing, effectively
reducing child loggers to symbolic names; assuming all loggers except the root
logger to be unconfigured, messages will be logged _only_ by the root logger.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To avoid circular import dependencies, avoid importing from *ANY*
# application-specific modules at the top-level -- excluding those explicitly
# known *NOT* to import from this module. Since all application-specific modules
# must *ALWAYS* be able to safely import from this module at any level, these
# circularities are best avoided here rather than elsewhere.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

import logging, sys, traceback
from betse.util.io.log.logenum import LogLevel
from betse.util.type import types
from betse.util.type.types import type_check, StrOrNoneTypes

# ....................{ GETTERS                            }....................
@type_check
def get(logger_name: StrOrNoneTypes = None) -> logging.Logger:
    '''
    Logger with the passed ``.``-delimited name, defaulting to the basename of
    the current process (e.g., ``betse``) implying the **global logger** (i.e.,
    the default application-wide logger).

    This function expects the :class:`LogConfig` class to have been previously
    instantiated, which globally configures logging.

    Parameters
    ----------
    logger_name : StrOrNoneTypes
        ``.``-delimited name of the logger to retrieve. By convention, logger
        names are typically that of the calling module (e.g., ``__name__``).
        However, since our default logger configuration already dynamically
        embeds the name of the calling module but *not* this logger into log
        records, an explicit logger name need *not* be passed in most cases.
        The exception to this heuristic is third-party dependencies whose
        package-specific loggers require configuration (e.g., to reduce the
        verbosity of messages logged by those loggers), in which case those
        loggers' names should be explicitly passed. Defaults to ``None``, in
        which case the global logger is retrieved.
    '''

    # Default the name of this logger to the name of the root logger.
    if logger_name is None:
        logger_name = logging.root.name

    # If this name is the empty string, this function would get the root logger.
    # Since this name being empty typically constitutes an implicit error rather
    # than an attempt to get the root logger, prevent this via an assertion.
    assert types.is_str_nonempty(logger_name), (
        types.assert_not_str_nonempty(logger_name, 'Logger name'))

    # Return this logger.
    return logging.getLogger(logger_name)

# ....................{ LOGGERS ~ banner                   }....................
@type_check
def log_banner(*args, **kwargs) -> None:
    '''
    Log a single-line banner described by the passed arguments as an
    informational message with the root logger.

    Since formatting this banner requires this banner's embedded title to be
    preformatted, this function *does* not accept the ``%``-style positional and
    keyword arguments accepted by most other logging functions (e.g.,
    :func:`log_info`).

    This function expects the :class:`LogConfig` class globally configuring
    logging to be instantiated as a singleton.

    Parameters
    ----------
    All passed arguments are passed as is to the
    :func:`betse.util.os.shell.shellstr.get_banner` function.
    '''

    # Avoid circular import dependencies.
    from betse.util.os.shell import shellstr

    # Single-line banner embedding the passed human-readable title.
    banner = shellstr.get_banner(*args, **kwargs)

    # Log this banner with level "INFO".
    log_info(banner)

# ....................{ LOGGERS ~ level                    }....................
@type_check
def log_levelled(message: str, level: LogLevel, *args, **kwargs) -> None:
    '''
    Log the passed message of the passed logging level (e.g.,
    :attr:`LogLevel.INFO`) with the root logger, formatted with the passed
    ``%``-style positional and keyword arguments.

    This function expects the :class:`LogConfig` class globally configuring
    logging to be instantiated as a singleton.

    Parameters
    ----------
    message : str
        Message to log containing zero or more ``%``-style format specifiers.
    level : LogLevel
        Logging level to log this message with (e.g., :attr:`LogLevel.INFO`).

    ALl remaining parameters are interpolated into the message according to the
    ``%``-style format specifiers embedded in this message.
    '''

    # The Logger.log() method accepts these parameters in the opposite order.
    logging.log(level, message, *args, **kwargs)


@type_check
def log_debug(message: str, *args, **kwargs) -> None:
    '''
    Log the passed debug message with the root logger, formatted with the passed
    ``%``-style positional and keyword arguments.

    This function expects the :class:`LogConfig` class globally configuring
    logging to be instantiated as a singleton.
    '''

    logging.debug(message, *args, **kwargs)


@type_check
def log_info(message: str, *args, **kwargs) -> None:
    '''
    Log the passed informational message with the root logger, formatted with
    the passed ``%``-style positional and keyword arguments.

    This function expects the :class:`LogConfig` class globally configuring
    logging to be instantiated as a singleton.
    '''

    logging.info(message, *args, **kwargs)


@type_check
def log_warning(message: str, *args, **kwargs) -> None:
    '''
    Log the passed warning message with the root logger, formatted with the
    passed ``%``-style positional and keyword arguments.

    This function expects the :class:`LogConfig` class globally configuring
    logging to be instantiated as a singleton.
    '''

    logging.warning(message, *args, **kwargs)


@type_check
def log_error(message: str, *args, **kwargs) -> None:
    '''
    Log the passed error message with the root logger, formatted with the
    passed ``%``-style positional and keyword arguments.

    This function expects the :class:`LogConfig` class globally configuring
    logging to be instantiated as a singleton.
    '''

    logging.error(message, *args, **kwargs)

# ....................{ LOGGERS ~ exception                }....................
@type_check
def log_exception(exception: Exception) -> None:
    '''
    Log the passed exception with the root logger.
    '''

    # While all loggers provide an exception() method for logging exceptions,
    # the output produced by these methods is in the same format as that
    # produced by the Python interpreter on uncaught exceptions. In order, this
    # is:
    #
    # * This exception's non-human-readable stack trace.
    # * This exception's human-readable error message.
    #
    # Since this format is (arguably) unreadable for non-developers, this
    # exception is reformatted for readability. Sadly, this precludes calling
    # our logger's exception() method.
    #
    # Attempt to...
    try:
        # Avoid circular import dependencies.
        from betse.util.io import stderrs
        from betse.util.io.error import errexception
        from betse.util.io.log import logconfig

        # Terse synopsis and verbose traceback for this exception.
        exc_synopsis, exc_traceback = errexception.get_metadata(exception)

        # Singleton logging configuration for the current Python process.
        log_config = logconfig.get()

        # If the end user requested that nothing be logged to disk, respect this
        # request by logging tracebacks to the error level and hence stderr.
        # (Avoid printing the synopsis already embedded in these tracebacks.)
        if log_config.file_level >= LogLevel.NONE:
            log_error(exc_traceback)
        # Else, the end user requested that at least something be logged to
        # disk. For debuggability, the logging level of the file handler is
        # temporarily decreased to the debug level, guaranteeing that tracebacks
        # are *ALWAYS* at least logged to disk rather than (possibly) discarded.
        # For readability, tracebacks are only logged to stderr if explicitly
        # requested by the end user.
        else:
            # If verbosity is disabled, output this synopsis to stderr;
            # else, tracebacks containing this synopsis are already
            # output to stderr by logging performed below.
            if not log_config.is_verbose:
                # Print this synopsis followed by a human-readable reference to
                # the current logfile.
                stderrs.output(
                    '{}\n\nFor details, see "{}".'.format(
                        exc_synopsis, log_config.filename))

            # Previous minimum level of messages to log to disk.
            log_config_file_level = log_config.file_level

            # Temporarily coerce this to the debug level, ensuring that
            # tracebacks are *ALWAYS* at least logged to disk.
            log_config.file_level = LogLevel.DEBUG

            # Attempt to...
            try:
                # Log tracebacks to the debug level and hence *NOT* stderr by
                # default, isolating tracebacks to disk. This is a Good Thing.
                # Tracebacks supply more detail than desired by typical users.
                log_debug(exc_traceback)
            # Revert to the previous level even if an exception is raised.
            finally:
                log_config.file_level = log_config_file_level
    # If this handling raises an exception, catch and print this exception
    # via the standard Python library, guaranteed not to raise exceptions.
    except Exception:
        # Header preceding the exception to be printed.
        exc_heading = 'log_exception() recursively raised exception:\n'

        # If the stderrs.output_exception() function exists, call that function.
        if 'stderrs' in locals():
            stderrs.output_exception(heading=exc_heading)
        # Else, something has gone horribly wrong. Defer to stock functionality
        # in the standard "traceback" module.
        else:
            print(exc_heading, file=sys.stderr)
            traceback.print_exc(file=sys.stderr)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **generator** (i.e., objects satisfying the standard generator API)
facilities.
'''

# ....................{ IMPORTS                            }....................
from betse.util.type.types import GeneratorType

# ....................{ GENERATORS                         }....................
def empty_generator() -> GeneratorType:
    '''
    Empty generator yielding... absolutely nothing.

    See Also
    ----------
    https://stackoverflow.com/a/36658865/2809027
        StackOverflow answer strongly inspiring this implementation.
    '''

    yield from ()



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
BETSE-specific test exception hierarchy.
'''

# ....................{ IMPORTS                            }....................
from abc import ABCMeta

# ....................{ EXCEPTIONS                         }....................
class BetseTestException(Exception, metaclass=ABCMeta):
    '''
    Abstract base class of all BETSE-specific test exceptions.
    '''
    pass


class BetseTestFixtureException(BetseTestException):
    '''
    Fixture-specific test exception.
    '''
    pass


class BetseTestHookException(BetseTestException):
    '''
    Hook-specific test exception.
    '''
    pass


class BetseTestParamException(BetseTestException):
    '''
    Parameter-specific test exception.
    '''
    pass



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2017 by Alexis Pietak & Cecil Curry
# See "LICENSE" for further details.

import time
import matplotlib.pyplot as plt
from betse.exceptions import BetseExceptionSimulation, BetseExceptionParameters
from betse.science import filehandling as fh
from betse.science.parameters import Parameters
from betse.science.plot.pipeline import plot_all
from betse.science.sim import Simulator
from betse.util.io.log import logs
from betse.util.path import files, paths
from betse import ignition, pathtree

# Initialize the current application before doing anything else.
ignition.init()
# at the moment only the default config file can be used
config_filename = pathtree.CONFIG_DEFAULT_FILENAME

# Validate and localize such filename.
files.die_unless_file(config_filename)
_config_filename = config_filename
_config_basename = paths.get_basename(_config_filename)


logs.log_info(
    'Running simulation with configuration file "{}".'.format(
        _config_basename))

start_time = time.time()  # get a start value for timing the simulation

p = Parameters(config_filename=_config_filename)  # create an instance of Parameters
p.set_time_profile(p.time_profile_sim)  # force the time profile to be initialize
p.run_sim = True  # set on the fly a boolean to let simulator know we're running a full simulation
sim = Simulator(p)  # create an instance of Simulator

if files.is_file(sim.savedInit):
    sim, cells, p_old = fh.loadSim(sim.savedInit)  # load the initialization from cache
    p.sim_ECM = cells.sim_ECM

    # check to ensure compatibility between original and present sim files:
    if p_old.config['general options'] != p.config['general options'] or\
            p_old.config['world options'] != p.config['world options']:
        # p_old.config['tissue profile definition'] != p.config['tissue profile definition']:
        raise BetseExceptionParameters(
                    'Important config file options are out of sync between '
                    'seed and this init/sim attempt! '
                    'Run "betse seed" again to match the current settings of '
                    'this config file.')

else:
    logs.log_info("No initialization file found to run this simulation!")

    raise BetseExceptionSimulation("Simulation terminated due to missing initialization. Please run "
                                       "an initialization and try again.")

# Reinitialize save and load directories in case params defines new ones
# for this sim.
sim.fileInit(p)

# Run and save the simulation to the cache.
sim.sim_info_report(cells, p)

sim.run_sim_core(cells, p)

logs.log_info(
    'The simulation took {} seconds to complete.'.format(
        round(time.time() - start_time, 2)))
logs.log_info(
    'When ready, close all of the figure windows to end the program.')

if p.turn_all_plots_off is False:
    # As colormaps are deleted from p prior to saving in sim, create a
    # fresh instance of Parameters.
    p = Parameters(config_filename=_config_filename)
    plot_all(cells, sim, p, plot_type='sim')
    plt.show()







#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level versioning facilities.
'''

# ....................{ IMPORTS                            }....................
import pkg_resources
from betse.util.type.types import type_check

# ....................{ TESTERS                            }....................
@type_check
def is_at_least(version1: str, version2: str) -> bool:
    '''
    ``True`` only if the version described by the first passed ``.``-delimited
    version specifier is at least as recent as the version described by the
    second such specifier.

    Parameters
    ----------
    version1: str
        First ``.``-delimited version specifier (e.g., ``2.0.1``) to test.
    version2: str
        Second ``.``-delimited version specifier (e.g., ``1.0.2``) to test.

    Returns
    ----------
    bool
        ``True`` only if the first version is as recent as the second version.

    See Also
    ----------
    https://stackoverflow.com/a/21065570/2809027
        StackOverflow overflow inspiring this implementation.
    '''

    # Setuptools. It's actually good for something.
    #
    # Note that there are *MANY* varying approaches for comparing version
    # specifiers from Python, including:
    #
    # * The packaging.version.parse() function, whose API resembles that of
    #   setuptools called below. In theory, the "packaging" package *SHOULD* be
    #   a setuptools dependency and hence always importable wherever the
    #   "pkg_resources" and "setuptools" packages are importable. In practice,
    #   this does *NOT* appear to be the case.
    # * The "distutils.version.LooseVersion" class, whose API resembles that of
    #   setuptools called below. However, whereas the latter API is PEP
    #   440-conformant, the latter API is not.
    #
    # Hence, the current approach.
    return (
        pkg_resources.parse_version(version1) >=
        pkg_resources.parse_version(version2)
    )



#!/usr/bin/env python3
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Abstract base classes of all timed event classes.
'''

# ....................{ IMPORTS                            }....................
from abc import ABCMeta  #, abstractmethod
from betse.util.type.types import type_check, NumericSimpleTypes

# ....................{ BASE                               }....................
class SimEventABC(object, metaclass=ABCMeta):
    '''
    Abstract base class of all timed event classes.

    Instances of this class parameterize an exogenous event (e.g., cell removal,
    voltage change) to be triggered at some time step(s) of the simulation.

    Attributes
    ----------
    _is_fired : bool
        ``True`` only if this event's :meth:`fire` method has already been
        called for a previous time step of the appropriate simulation phase.
    '''

    # ..................{ INITIALIZERS                       }..................
    @type_check
    def __init__(self) -> None:
        '''
        Initialize this event.
        '''

        # Classify all passed parameters.
        self._is_fired = False

    # ..................{ PROPERTIES                         }..................
    # Read-only properties, preventing callers from setting these attributes.

    @property
    def is_fired(self) -> bool:
        '''
        ``True`` only if this event's :meth:`fire` method has already been
        called for a previous time step of the appropriate simulation phase.
        '''

        return self._is_fired

    # ..................{ FIRERS                             }..................
    @type_check
    def fire(self) -> None:
        '''
        Apply this event.

        This method notes this event to have been applied, ensuring the
        :meth:`is_fired` property to subsequently report ``True``.
        '''

        self._is_fired = True

# ....................{ SUBCLASSES                         }....................
class SimEventSpikeABC(SimEventABC):
    '''
    Abstract base class of all classes describing simulation events occurring at
    only a single time step (rather than over a range of time steps).

    Attributes
    ----------
    _time_step : NumericSimpleTypes
        Time step in seconds (s) at which to trigger this action.
    '''

    # ..................{ INITIALIZERS                       }..................
    @type_check
    def __init__(self, time_step: NumericSimpleTypes) -> None:
        '''
        Initialize this event.

        Parameters
        ----------
        time_step : NumericSimpleTypes
            Time step in seconds (s) at which to trigger this action.
        '''

        # Initialize our superclass.
        super().__init__()

        # Classify all passed parameters.
        self._time_step = time_step


class SimEventPulseABC(SimEventABC):
    '''
    Abstract base class of all classes describing simulation events occurring
    over a range of time steps (rather than at only a single time step).

    Attributes
    ----------
    start_time : NumericSimpleTypes
        Time step (s) at which to begin triggering this event.
    stop_time : NumericSimpleTypes
        Time step (s) at which to cease triggering this event.
    step_rate : NumericSimpleTypes
        Slope of the pair of step functions guaranteeing smooth continuity
        between the background function and this event. Each step function is
        the mirror image of the other reflected across the Y axis. These are:
        * A "step" up from the background function to this event, whose
          mid-point is centered at ``start_time``.
        * A "step" down from this event back to the background function, whose
          mid-point is centered at ``stop_time``.
        If the background function is time-dependent, this slope is a **rate**
        (i.e., change over time). For the :class:`SimEventPulseVoltage` subclass, for
        example, this is the rate in voltage per seconds (V/s) at which:
        * The background voltage is first increased to the peak voltage.
        * The peak voltage is later decreased to the background voltage.
    '''

    # ..................{ INITIALIZERS                       }..................
    @type_check
    def __init__(
        self,
        start_time: NumericSimpleTypes,
        stop_time: NumericSimpleTypes,
        step_rate: NumericSimpleTypes,
    ) -> None:

        # Initialize our superclass.
        super().__init__()

        # Classify all passed parameters.
        self.start_time = start_time
        self.stop_time = stop_time
        self.step_rate = step_rate



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level support facilities for Pillow, the Python Image Library
(PIL)-compatible fork underlying image I/O performed by this application.
'''

#FIXME: Revisit imageio when the following feature request is resolved in full:
#    https://github.com/imageio/imageio/issues/289

# ....................{ IMPORTS                            }....................
from PIL import Image
from betse.util.io.log import logs
from betse.util.path import pathnames
from betse.util.type.decorator.decmemo import func_cached
from betse.util.type.types import SetType  # type_check,

# ....................{ GETTERS                            }....................
@func_cached
def get_filetypes() -> SetType:
    '''
    Set of all image filetypes supported by the current version of Pillow.

    For generality, these filetypes are *not* prefixed by a ``.`` delimiter.

    Examples
    ----------
        >>> from betse.lib.pil import pils
        >>> pils.get_filetypes()
        {'flc', 'bmp', 'ppm', 'webp', 'j2k', 'jpf', 'jpe', 'pcd'}
    '''

    # Initialize Pillow if uninitialized.
    #
    # If Pillow is uninitialized, the "Image.EXTENSION" dictionary is empty.
    # Since the "betse.ignition" submodule already initializes Pillow,
    # explicitly doing so here should typically *NOT* be necessary. Since this
    # getter could technically be called from global scope prior to the
    # initialization performed by "betse.ignition" *AND* since this
    # initialization efficiently reduces to a noop if unnecessary, manually
    # initializing Pillow here is cost-free. (Cost-free is the way to be.)
    init()

    # Return a set of...
    return set(
        # This filetype strips of this prefixing ".".
        pathnames.undot_filetype(filetype_dotted)
        # For each "."-prefixed filetype supported by Pillow...
        for filetype_dotted in Image.EXTENSION.keys()
    )

# ....................{ ENUMERATIONS                       }....................
def init() -> None:
    '''
    Initialize Pillow if uninitialized *or* reduce to a noop otherwise (i.e., if
    Pillow is already initialized).
    '''

    # Log this initialization.
    logs.log_debug('Initializing Pillow...')

    # If Pillow is already initialized, this function internally reduces to an
    # efficient noop in the expected manner (e.g., by accessing a private
    # boolean global).
    Image.init()



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Utility functions simplifying access to various :mod:`pytest` subpackages,
submodules, and attributes unclassifiable with a more specific designation.
'''

# ....................{ IMPORTS                            }....................
from betse.util.type.types import ClassType, ModuleType

# ....................{ GETTERS                            }....................
def get_pytest_fixtures_submodule() -> ModuleType:
    '''
    Version-specific private :mod:`pytest` submodule providing fixture classes.

    Although private, this submodule publishes classes of public interest.
    Unfortunately, as this submodule *is* private, its fully-qualified name has
    changed across :mod:`pytest` versions. This function guaranteeably retrieves
    this submodule regardless of version.

    Specifically, under:

    * :mod:`pytest` >= 3.0.0, the private :mod:`_pytest.fixtures` submodule is
      imported and returned.
    * :mod:`pytest` < 3.0.0, the private :mod:`_pytest.python` submodule is
      imported and returned.
    '''

    # Attempt to import and return the newer "pytest" fixture module.
    try:
        from _pytest import fixtures as pytest_fixtures
    # Failing that, fallback to importing and returning the older such module.
    except ImportError:
        from _pytest import python as pytest_fixtures

    # In either case, return this module.
    return pytest_fixtures

# ....................{ GETTERS ~ type                     }....................
def get_fixture_lookup_error_type() -> ClassType:
    '''
    Class of all :class:`FixtureLookupError` exceptions raised by :mod:`pytest`.

    Specifically, under:

    * :mod:`pytest` >= 3.0.0, the private
      :class:`_pytest.fixtures.FixtureLookupError` class is imported and
      returned.
    * :mod:`pytest` < 3.0.0, the private
      :class:`_pytest.python.FixtureLookupError` class is imported and returned.
    '''

    # Version-specific private `pytest` submodule providing fixture classes.
    pytest_fixtures_submodule = get_pytest_fixtures_submodule()

    # Return this exception class defined by this submodule.
    return pytest_fixtures_submodule.FixtureLookupError



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2017 by Alexis Pietak & Cecil Curry
# See "LICENSE" for further details.

import time

import matplotlib.pyplot as plt
from betse.exceptions import BetseExceptionSimulation, BetseExceptionParameters
from betse.science import filehandling as fh
from betse.science.cells import Cells
from betse.science.parameters import Parameters
from betse.science.plot.pipeline import plot_all
from betse.science.sim import Simulator
from betse.util.io.log import logs
from betse.util.path import files, paths
from betse import ignition, pathtree

# Initialize the current application before doing anything else.
ignition.init()
# at the moment only the default config file can be used
config_filename = pathtree.CONFIG_DEFAULT_FILENAME

# Validate and localize such filename.
files.die_unless_file(config_filename)
_config_filename = config_filename
_config_basename = paths.get_basename(_config_filename)

logs.log_info(
    'Initializing simulation with configuration file "{}".'.format(
        _config_basename))

start_time = time.time()  # get a start value for timing the simulation

p = Parameters(config_filename=_config_filename)  # create an instance of Parameters
p.set_time_profile(p.time_profile_init)  # force the time profile to be initialize
p.run_sim = False  # let the simulator know we're just running an initialization

# cells, _ = fh.loadSim(cells.savedWorld)
cells = Cells(p)  # create an instance of world

if files.is_file(cells.savedWorld):
    cells, p_old = fh.loadWorld(cells.savedWorld)  # load the simulation from cache
    logs.log_info('Cell cluster loaded.')

    # check to ensure compatibility between original and present sim files:
    if p_old.config['general options'] != p.config['general options'] or\
            p_old.config['world options'] != p.config['world options']:
        # p_old.config['tissue profile definition'] != p.config['tissue profile definition']:
        raise BetseExceptionParameters(
                    'Important config file options are out of sync between '
                    'seed and this init/sim attempt! '
                    'Run "betse seed" again to match the current settings of '
                    'this config file.')

else:
    logs.log_info("Ooops! No such cell cluster file found to load!")

    raise BetseExceptionSimulation(
            "Run terminated due to missing seed.\n"
            "Please run 'betse seed' to try again.")

sim = Simulator(p)  # create an instance of Simulator
sim.run_sim = False

# Initialize simulation data structures, run, and save simulation phase
sim.baseInit_all(cells, p)
sim.sim_info_report(cells, p)
sim.run_sim_core(cells, p)

logs.log_info('Initialization run complete!')
logs.log_info(
    'The initialization took {} seconds to complete.'.format(
        round(time.time() - start_time, 2)))

if p.turn_all_plots_off is False:
    # as colormaps are deleted from p prior to saving in sim, create a fresh instance of Parameters:
    p = Parameters(config_filename=_config_filename)

    logs.log_info(
        'When ready, close all of the figure windows to proceed with '
        'scheduled simulation runs.')

    plot_all(cells, sim, p, plot_type='init')
    plt.show()


#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

#FIXME: Executables output under OS X and Windows pretty much *MUST* be signed.
#This looks to be fairly trivial under Windows. OS X, however, is another kettle
#of hideous fish. In any case, everyone else has already solved this, so we just
#need to leverage the following detailed recipes:
#
#* https://github.com/pyinstaller/pyinstaller/wiki/Recipe-Win-Code-Signing
#* https://github.com/pyinstaller/pyinstaller/wiki/Recipe-OSX-Code-Signing

#FIXME: Embed Windows-specific version metadata in such executables. This is a
#fairly bizarre process, which we've documented in "pyinstaller.yaml". It's
#hardly crucial for now, but will be important at some point.
#FIXME: Embed OS X-specific version metadata in such executables. We'll want to
#detect whether the current OS is OS X and, if so, manually overwrite the
#autogenerated "myapp.app/Contents/Info.plist" file with one of our own
#devising. Not terribly arduous... in theory.

#FIXME: Embed ".ico"-suffixed icon files in such executables. PyInstaller
#provides simple CLI options for this; we simply need to create such icons.
#Contemplating mascots, how about the ever-contemplative BETSE cow?
#
#Sadly, the icon formats required by OS X and Windows appear to conflict.
#Windows icon files have filetype ".ico" (and appear to support only one
#embedded icon), whereas OS X icon files have filetype ".icns" (and appear to
#support multiple embedded icons). To compound matters, "pyinstaller" provides
#only one option "--icon" for both, probably implying that we'll need to
#dynamically detect whether the current system is OS X or Windows and respond
#accordingly (i.e., by passing the appropriate system-specific icon file).

#FIXME: Contribute back to the community. Contemplate a stackoverflow answer.
#(We believe we may have espied an unanswered question asking about query words
#"pyinstaller setuptools integration". Huzzah!) We should note that PyInstaller
#will probably be unable to find the imports of setuptools-installed scripts,
#due to the obfuscatory nature of such scripts. See the following for a
#reasonable solution:
#    https://github.com/pyinstaller/pyinstaller/wiki/Recipe-Setuptools-Entry-Point

'''
BETSE-specific `freeze` subcommands for `setuptools`.
'''

# ....................{ IMPORTS                            }....................
import os
from abc import ABCMeta, abstractmethod
from distutils.errors import DistutilsExecError
from os import path
from pkg_resources import EntryPoint
from betse_setup import buputil
from setuptools import Command

# ....................{ COMMANDS                           }....................
def add_setup_commands(metadata: dict, setup_options: dict) -> None:
    '''
    Add `freeze` subcommands to the passed dictionary of `setuptools` options.
    '''

    buputil.add_setup_command_classes(
        metadata, setup_options, freeze_dir, freeze_file)

# ....................{ CLASSES ~ base                     }....................
class freeze(Command, metaclass = ABCMeta):
    '''
    Abstract command class creating either one platform-specific executable file
    _or_ one directory containing such file in the top-level `dist` directory
    for each previously installed script of the current application.

    Each such file is created by running the external `pyinstaller` command with
    sane command-line arguments. Since this command does _not_ support
    **cross-bundling** (i.e., creation of executable files for operating systems
    other than the current), these files apply _only_ to the current system.
    Specifically:

    * Under Linux, these files will be ELF (Executable and Linkable Format)
      binaries.
    * Under OS X, these files will be conventional `.app`-suffixed directories.
      (Of course, that's not a file. So sue us.)
    * Under Windows, such files will be conventional `.exe`-suffixed binaries.

    Attributes
    ----------
    clean : optional[bool]
        `True` only if the user passed the `--clean` option to the current
        `setuptools` subcommand. Defaults to `True`, for safety.
    debug : optional[bool]
        `True` only if the user passed the `--debug` option to the current
        `setuptools` subcommand. Defaults to `False`.
    dist_dir : optional[str]
        Absolute or relative path of the output directory to which frozen
        executables (as either files or directories) are written. Defaults to
        the relative path `freeze/dist/`.
    install_dir : str
        Absolute path of the directory to which our wrapper scripts were
        previously installed.
    _pyinstaller_args : list
        List of all shell words of the PyInstaller command to be run.
    _pyinstaller_spec_filename : str
        Relative path of our input PyInstaller custom spec file, freezing
        platform-agnostic wrapper scripts into platform-specific executables.
    _pyinstaller_hooks_dirname : str
        Relative path of our input PyInstaller custom hooks subdirectory.
    '''

    # ..................{ ATTRIBUTES                         }..................
    EXCEPTION_ADVICE = (
        'Consider running either:\n'
        '\tsudo python3 setup.py install\n'
        '\tsudo python3 setup.py symlink')
    '''
    Human-readable advice appended to exception messages raised by this class.
    '''


    user_options = [
        ('clean', None,
         'clean PyInstaller cache of temporary paths before building'),
        ('debug', None,
         'print debug messages during PyInstaller bootloader startup'),
        ('dist-dir=', None,
         'absolute or relative path of the output directory to which executables will be frozen'),
    ]
    '''
    List of 3-tuples specifying command-line options accepted by this command.

    For each such option, an attribute of the same name as this option's long
    form _must_ be explicitly initialized in the `initialize_options()` method.
    `setuptools` fails to recognize options for which this is _not_ the case.
    (You fail a simple sanity check yet again, `setuptools`.)

    See Also
    ----------
    http://ilostmynotes.blogspot.ca/2009/04/python-distutils-installer-and.html
        Inarguably, the best (albeit unofficial) documentation on this list.
    '''

    # ..................{ SUPERCLASS                         }..................
    def initialize_options(self):
        '''
        Declare option-specific attributes subsequently initialized by
        `finalize_options()`.

        If this function is _not_ defined, the default implementation of this
        method raises an inscrutable `distutils` exception. If these attributes
        are _not_ declared, the subsequent call to
        `self.set_undefined_options()` raises an inscrutable `setuptools`
        exception. (This is terrible. So much hate.)
        '''
        #FIXME: To circumvent PyInstaller caching issues, we currently force
        #"self.clean = True". Once PyInstaller caching is sufficiently reliable
        #to reasonably permit reuse of cached metadata, revert this back to
        #"self.clean = False".

        # Option-specific public attributes. For each option declared by the
        # "user_options" list above, a public attribute of the same name as this
        # option's long form *MUST* be initialized here to its default value.
        self.clean = True
        self.debug = False
        self.dist_dir = None

        # setuptools-specific public attributes.
        self.install_dir = None

        # Custom private attributes.
        self._pyinstaller_args = None
        self._pyinstaller_spec_filename = None
        self._pyinstaller_hooks_dirname = None


    def finalize_options(self):
        '''
        Default undefined command-specific options to the options passed to the
        current parent command if any (e.g., `symlink`).
        '''

        # Copy attributes from a temporarily instantiated "symlink" object into
        # the current object under different attribute names.
        self.set_undefined_options(
            'symlink', ('install_scripts', 'install_dir'))


    def run(self):
        '''
        Run the current command and all subcommands thereof.
        '''

        # Construct the PyInstaller command to be run.
        self._init_pyinstaller_command()

        # Run this command for each entry point.
        self._run_pyinstaller_commands()


    # ..................{ INITIALIZERS                       }..................
    def _init_pyinstaller_command(self) -> None:
        '''
        Initialize the list of all shell words of the PyInstaller command to be
        run.
        '''

        # Relative path of the top-level PyInstaller directory.
        pyinstaller_dirname = 'freeze'

        # Relative path of the PyInstaller spec file converting such
        # platform-independent script into a platform-specific executable.
        self._pyinstaller_spec_filename = path.join(
            pyinstaller_dirname, '.spec')

        # If the frozen executable directory was *NOT* explicitly passed on the
        # command-line, default to a subdirectory of this top-level directory.
        if self.dist_dir is None:
            self.dist_dir = path.join(pyinstaller_dirname, 'dist')
        # Else, canonicalize the passed directory.
        else:
            self.dist_dir = buputil.get_path_canonicalized(self.dist_dir)
        assert isinstance(self.dist_dir, str), (
            '"{}" not a string.'.format(self.dist_dir))

        # Relative path of the input hooks subdirectory.
        self._pyinstaller_hooks_dirname = path.join(
            pyinstaller_dirname, 'hooks')

        # Relative path of the intermediate build subdirectory.
        pyinstaller_work_dirname = path.join(pyinstaller_dirname, 'build')

        # Create such hooks subdirectory if not found, as failing to do so
        # will induce fatal PyInstaller errors.
        buputil.make_dir_unless_found(self._pyinstaller_hooks_dirname)

        # List of all shell words of the PyInstaller command to be run, starting
        # with the basename of this command.
        self._pyinstaller_args = []

        # Append all PyInstaller command options common to running such command
        # for both reuse and regeneration of spec files. (Most such options are
        # specific to the latter only and hence omitted.)
        self._pyinstaller_args = [
            # Overwrite existing output paths under the "dist/" subdirectory
            # without confirmation, the default behaviour.
            '--noconfirm',

            # Non-default PyInstaller directories.
            '--workpath=' + buputil.shell_quote(pyinstaller_work_dirname),
            '--distpath=' + buputil.shell_quote(self.dist_dir),

            # Non-default log level.
            # '--log-level=DEBUG',
            '--log-level=INFO',
        ]

        # Forward all custom boolean options passed by the user to the current
        # setuptools command (e.g., "--clean") to the "pyinstaller" command.
        if self.clean:
            self._pyinstaller_args.append('--clean')
        if self.debug:
            self._pyinstaller_args.extend((
                '--debug',

                # UPX-based compression uselessly consumes non-trivial time
                # (especially under Windows, where process creation is fairly
                # heavyweight) when freezing debug binaries. To optimize and
                # simplify debugging, such compression is disabled.
                '--noupx',
            ))
            buputil.output_warning(
                'Enabling bootloader debug messages.')
            buputil.output_warning(
                'Disabling UPX-based compression.')
        # If *NOT* debugging and UPX is *NOT* found, print a non-fatal warning.
        # While optional, freezing in the absence of UPX produces uncompressed
        # and hence considerably larger executables.
        elif not buputil.is_pathable('upx'):
            buputil.output_warning(
                'UPX not installed or "upx" not in the current ${PATH}.')
            buputil.output_warning(
                'Frozen binaries will *NOT* be compressed.')

    # ..................{ RUNNERS                            }..................
    def _run_pyinstaller_commands(self):
        '''
        Run the PyInstaller command previously constructed by the
        `_init_pyinstaller_command()` method for each entry point.
        '''

        # True if the current distribution has at least one entry point.
        is_entry_point = False

        # Freeze each previously installed script wrapper.
        for script_basename, script_type, entry_point in\
            buputil.command_entry_points(self):
            # Note at least one entry point to be installed.
            is_entry_point = True

            #FIXME: We went to a great deal of trouble to implement this method.
            #Why don't we call it anymore, again?

            # Validate this wrapper's entry point.
            # freeze._check_entry_point(entry_point)

            # Relative path of the output frozen executable file or directory,
            # created by stripping the suffixing ".exe" filetype on Windows.
            frozen_pathname = path.join(
                self.dist_dir,
                buputil.get_path_sans_filetype(script_basename))

            # If cleaning *AND* this path exists, remove this path before
            # validating this path. Why? This path could be an existing file
            # and the current command freezing to a directory (or vice versa),
            # in which case subsequent validation would raise an exception.
            if self.clean and buputil.is_path(frozen_pathname):
                buputil.remove_path(frozen_pathname)

            # Validate this path.
            self._check_frozen_path(frozen_pathname)

            # Set all environment variables used to communicate with the BETSE-
            # specific PyInstaller specification file run below.
            self._set_environment_variables(
                script_basename, script_type, entry_point)

            # Run this PyInstaller command for this entry point.
            self._run_pyinstaller_command(
                script_basename, script_type, entry_point)

            # Report these results to the user.
            frozen_pathtype = (
                'directory' if buputil.is_dir(frozen_pathname) else 'file')
            print('Froze {} "{}".\n'.format(frozen_pathtype, frozen_pathname))

            #FIXME: Excise when beginning GUI work.
            break

        # If no entry points are registered for the current distribution, raise
        # an exception.
        if not is_entry_point:
            raise DistutilsExecError(
                'No entry points found. {}'.format(freeze.EXCEPTION_ADVICE))


    #FIXME: Refactor to:
    #
    #1. Set global variables of this module rather than environment variables of
    #   the current Python process.
    #2. Refactor our spec file to import such global variables from this module.
    #
    #Since we now run PyInstaller via import in the current Python process
    #rather than as an external command in a different Python process, this
    #should be feasible.
    def _set_environment_variables(
        self, script_basename: str, script_type: str, entry_point: str) -> None:
        '''
        Set all environment variables used to communicate with the BETSE-
        specific PyInstaller specification file run in a separate process, given
        the passed arguments yielded by the `command_entry_points()` generator.

        While hardly ideal, PyInstaller appears to provide no other means of
        communicating with such file.
        '''

        # Absolute path of the entry module.
        #
        # This module's relative path to the top-level project directory is
        # obtained by converting the entry point specifier defined by "setup.py"
        # for the current entry point (e.g., "betse.gui.guicli:main") into a
        # platform-specific path. Sadly, setuptools provides no cross-platform
        # API for reliably obtaining the absolute path of the corresponding
        # script wrapper. Even if it did, such path would be of little use under
        # POSIX-incompatible platforms (e.g., Windows), where these wrappers are
        # binary blobs rather than valid Python scripts.
        #
        # Instead, we reverse-engineer the desired path via brute-force path
        # manipulation. Thus burns out another tawdry piece of my soul.
        module_filename = path.join(
            buputil.get_project_dirname(),
            entry_point.module_name.replace('.', path.sep) + '.py')

        # Ensure such module exists.
        buputil.die_unless_file(module_filename)

        # Such path.
        os.environ['__FREEZE_MODULE_FILENAME'] = module_filename

        # Whether to freeze in "one-file" or "one-directory" mode.
        os.environ['__FREEZE_MODE'] = self._get_freeze_mode()

        # Whether to freeze a CLI- or GUI-based application.
        os.environ['__FREEZE_INTERFACE_TYPE'] = script_type


    def _run_pyinstaller_command(
        self,
        script_basename: str,
        script_type: str,
        entry_point: 'EntryPoint',
    ) -> None:
        '''
        Run the currently configured PyInstaller command for the passed entry
        point's script wrapper.

        Attributes
        ----------
        script_basename : str
            Basename of the executable wrapper script running this entry point.
        script_type : str
            Type of the executable wrapper script running this entry point,
            guaranteed to be either:
            * `console` if this script is console-specific.
            * `gui` otherwise..
        entry_point : EntryPoint
            `EntryPoint` object, whose attributes specify the module to be
            imported and function to be run by this script.
        '''

        # If this spec exists, instruct PyInstaller to reuse rather than
        # recreate this file, thus preserving edits to this file.
        if buputil.is_file(self._pyinstaller_spec_filename):
            print('Reusing spec file "{}".'.format(
                self._pyinstaller_spec_filename))

            # Append the relative path of this spec file.
            self._pyinstaller_args.append(
                buputil.shell_quote(self._pyinstaller_spec_filename))

            # Freeze this script with this spec file.
            self._run_pyinstaller_imported()
        # Else, instruct PyInstaller to (re)create this spec file.
        else:
            # Absolute path of the directory containing this files.
            pyinstaller_spec_dirname = buputil.get_path_dirname(
                self._pyinstaller_spec_filename)

            # Absolute path of the current script wrapper.
            script_filename = path.join(
                self.install_dir, script_basename)
            buputil.die_unless_file(
                script_filename, 'File "{}" not found. {}'.format(
                    script_filename, freeze.EXCEPTION_ADVICE))

            # Inform the user of this action *AFTER* the above validation.
            # Since specification files should typically be reused rather
            # than regenerated, do so as a non-fatal warning.
            buputil.output_warning(
                'Generating spec file "{}".'.format(
                    self._pyinstaller_spec_filename))

            # Append all options specific to spec file generation.
            self._pyinstaller_args.extend([
                # If this is a console script, configure standard input and
                # output for console handling; else, do *NOT* and, if the
                # current operating system is OS X, generate an ".app"-suffixed
                # application bundle rather than a customary executable.
                '--console' if script_type == 'console' else '--windowed',

                # Non-default PyInstaller directories.
                '--additional-hooks-dir=' + buputil.shell_quote(
                    self._pyinstaller_hooks_dirname),
                '--specpath=' + buputil.shell_quote(pyinstaller_spec_dirname),
            ])

            # Append all subclass-specific options.
            self._pyinstaller_args.extend(self._get_pyinstaller_options())

            # Append the absolute path of this script.
            self._pyinstaller_args.append(buputil.shell_quote(script_filename))

            # Freeze this script and generate a spec file.
            self._run_pyinstaller_imported()

            # Absolute path of this file.
            script_spec_filename = path.join(
                pyinstaller_spec_dirname, script_basename + '.spec')

            # Rename this file to have the basename expected by the prior
            # conditional on the next invocation of this setuptools command.
            #
            # Note that "pyinstaller" accepts an option "--name" permitting
            # the basename of this file to be specified prior to generating
            # this file. Unfortunately, this option *ALSO* specifies the
            # basename of the generated executable. While the former is reliably
            # renamable, the former is *NOT* (e.g., due to code signing). Hence,
            # this file is manually renamed without passing this option.
            buputil.move_file(
                script_spec_filename, self._pyinstaller_spec_filename)


    def _run_pyinstaller_imported(self) -> None:
        '''
        Run the currently configured PyInstaller command within the current
        Python process -- rather than as an external command in a different
        Python process.

        This function imports and executes PyInstaller's CLI implementation in
        the current Python process, circumventing the inevitable complications
        that arise when running PyInstaller as an external command.
        '''

        # PyInstaller's top-level "__main__" module, providing programmatic
        # access to its CLI implementation.
        pyinstaller_main = buputil.import_module(
            'PyInstaller.__main__', exception_message=(
            'PyInstaller not installed under the current Python interpreter.'))

        # Run PyInstaller and propagate its exit status as ours to the caller.
        print('Running PyInstaller with arguments: {}'.format(
            self._pyinstaller_args))
        buputil.exit_with_status(
            pyinstaller_main.run(pyi_args=self._pyinstaller_args))

    # ..................{ CLASS                              }..................
    @classmethod
    def _check_entry_point(entry_point: EntryPoint):
        '''
        Validate the passed entry point, describing the current script wrapper
        to be frozen.
        '''
        assert isinstance(entry_point, EntryPoint), (
            '"{}" not an entry point.'.format(entry_point))

        # If this entry module is unimportable, raise an exception.
        if not buputil.is_module(entry_point.module_name):
            raise ImportError(
                'Entry module "{}" unimportable. {}'.format(
                entry_point.module_name, freeze.EXCEPTION_ADVICE))

        # If this entry module's basename is *NOT* "__main__", print a
        # non-fatal warning. For unknown reasons, attempting to freeze
        # customary modules usually causes the frozen executable to reduce
        # to a noop (i.e., silently do nothing).
        if entry_point.module_name.split('.') != '__main__':
            buputil.output_warning(
                'Entry module "{}" basename not "__main__".'.format(
                entry_point.module_name))

        # If this entry module has no entry function, print a non-fatal
        # warning. For unknown reasons, attempting to freeze without an
        # entry function usually causes the frozen executable to reduce to a
        # noop (i.e., silently do nothing).
        if not len(entry_point.attrs):
            buputil.output_warning(
                'Entry module "{}" entry function undefined.'.format(
                entry_point.module_name))

    # ..................{ SUBCLASS                           }..................
    @abstractmethod
    def _check_frozen_path(self, frozen_pathname: str) -> None:
        '''
        Validate the passed path to which PyInstaller will subsequently write
        the output frozen executable file or directory for the current input
        Python script (in a subclass-specific manner).
        '''
        pass

    @abstractmethod
    def _get_freeze_mode(self) -> str:
        '''
        Get a string constant specific to this subclass.

        This constant should be `file` when freezing in "one-file" mode and
        `dir` when freezing in "one-directory" mode. The BETSE-specific
        `__FREEZE_MODE` environment variable will be set to this constant,
        informing the BETSE-specific PyInstaller specification file run in a
        separate process of which mode to freeze in.
        '''
        pass

    @abstractmethod
    def _get_pyinstaller_options(self) -> list:
        '''
        Get a list of subclass-specific command-line options to be passed to the
        external `pyinstaller` command, when running such command in the absence
        of a previously generated spec file.

        When run with a previously generated spec file, such command effectively
        accepts *no* such options. Hence, this method is only called when no
        such file exists for the current script wrapper to be frozen.
        '''
        pass

# ....................{ CLASSES ~ sub                      }....................
class freeze_dir(freeze):
    '''
    Create one platform-specific executable file in one subdirectory of the
    top-level `dist` directory for each previously installed script for the
    current application.

    See Also
    ----------
    freeze
        For further details.
    '''

    description = (
        'freeze each installed entry point to a directory containing '
        'a platform-specific executable and all requisite dependencies'
    )
    '''
    Command description printed on running `./setup.py --help-commands`.
    '''

    def _check_frozen_path(self, frozen_pathname: str) -> None:
        '''
        Validate that the directory to be generated is *not* an existing file
        (e.g., due to a prior run of the `freeze_file` command).

        Additionally, if such directory exists *and* the user passed option
        `--clean`, such directory will be recursively deleted in a safe manner
        (e.g., *not* following symbolic links outside such directory).
        '''
        buputil.die_unless_dir_or_not_found(frozen_pathname)

    def _get_freeze_mode(self) -> str:
        return 'dir'

    def _get_pyinstaller_options(self) -> list:
        return [
            '--onedir',
        ]


class freeze_file(freeze):
    '''
    Create one platform-specific executable file in the top-level `dist`
    directory for each previously installed script for the current application.

    See Also
    ----------
    freeze
        For further details.
    '''


    description = (
        'freeze each installed entry point to a platform-specific executable'
    )
    '''
    Command description printed on running `./setup.py --help-commands`.
    '''


    def _check_frozen_path(self, frozen_pathname: str) -> None:
        '''
        Validate that the file to be generated is *not* an existing directory
        (e.g., due to a prior run of the "freeze_dir" command).

        Additionally, if such file exists *and* the user passed option
        `--clean`, such file will be deleted.
        '''

        buputil.die_unless_file_or_not_found(frozen_pathname)


    def _get_freeze_mode(self) -> str:

        return 'file'

    def _get_pyinstaller_options(self) -> list:

        return [
            '--onefile',
        ]



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level **simulation pipeline** (i.e., sequence of similar simulation
activities to be iteratively run) functionality.
'''

# ....................{ IMPORTS                            }....................
from abc import ABCMeta
from betse.exceptions import (
    BetseSimPipeException, BetseSimPipeRunnerUnsatisfiedException)
from betse.lib.yaml.abc.yamllistabc import YamlListItemTypedABC
from betse.science.phase.pipe.piperun import SimPipeRunner
from betse.science.phase.phasecls import SimPhase
from betse.util.io.log import logs
from betse.util.type.decorator.deccls import abstractproperty
from betse.util.type.obj import objects
from betse.util.type.text import strs
from betse.util.type.types import type_check, GeneratorType, IterableTypes

# ....................{ SUPERCLASSES                       }....................
class SimPipeABC(object, metaclass=ABCMeta):
    '''
    Abstract base class of all subclasses running a **simulation pipeline**
    (i.e., sequence of similar simulation activities to be iteratively run).

    This class implements the infrastructure for iteratively running all
    currently enabled simulation activities (referred to as "runners")
    comprising the pipeline defined by this subclass, either in parallel *or* in
    series.

    Design
    ----------
    Each subclass is expected to define:

    * One or more public methods with names prefixed by ``run_`` (e.g.,
      ``run_voltage_intra``). For each such method, the name of that method
      excluding the prefix ``run_`` is the name of that method's runner (e.g.,
      ``voltage_intra`` for the method name ``run_voltage_intra``). Each such
      method should:
      * Accept exactly one parameter: ``self``.
      * Return nothing (i.e.,``None``).
    * The abstract :meth:`_runners_conf` property returning a sequence of the
      names of all runners currently enabled by this pipeline (e.g.,
      ``['voltage_intra', 'ion_hydrogen', 'electric_total']``).

    The :meth:`run` method defined by this base class then dynamically
    implements this pipeline by iterating over the :meth:`_runners_conf` property
    and, for each enabled runner, calling that runner's method.

    Attributes (Private)
    ----------
    _phase : SimPhase
        Current simulation phase.

    Attributes (Private: Labels)
    ----------
    _label_singular_lowercase : str
        Human-readable lowercase singular noun synopsizing the type of runners
        implemented by this subclass (e.g., ``animation``, ``plot``).
    _label_singular_uppercase : str
        Human-readable singular noun whose first character is uppercase and all
        remaining characters lowercase (e.g., ``Animation``, ``Plot``).
    _label_plural_lowercase : str
        Human-readable lowercase plural noun synopsizing the type of runners
        implemented by this subclass (e.g., ``animations``, ``plots``).
    '''

    # ..................{ CONSTANTS                          }..................
    # Ideally, the following class constants would instead be implemented as
    # class properties. Unfortunately, there exists no @classproperty decorator
    # and no feasible means of implementing that decorator thanks to subtle
    # idiosyncracies in the design of Python's data descriptor protocol. While
    # class properties can technically be manually implemented by a metaclass
    # defining these properties, doing so is cumbersome and quite discouraged.

    _RUNNER_METHOD_NAME_PREFIX = 'run_'
    '''
    Substring prefixing the name of each runner defined by this pipeline.

    This class variable is intended to be overridden by subclasses desiring a
    less ambiguous prefix (e.g., ``export_`` for export pipelines).
    '''

    # ..................{ STATIC ~ iterators                 }..................
    @classmethod
    def iter_runners(cls) -> GeneratorType:
        '''
        Generator yielding the 2-tuple ``(runner_name, runner)`` for each
        **runner** (i.e., :class:`SimPipeRunner` instance produced by the
        :func:`piperunner` decorator decorating this runner's method)
        defined by this pipeline subclass.

        This generator excludes all methods defined by this pipeline subclass
        *not* decorated by this decorator.

        Yields
        ----------
        (str, SimPipeRunner)
            2-tuple ``(runner_name, runner)`` where:
            * ``runner_name`` is the name of this runner's underlying method
              excluding the substring :attr:`_RUNNER_METHOD_NAME_PREFIX`
              prefixing this name.
            * ``runner`` is each runner's :class:`SimPipeRunner` instance.
        '''

        # Return a generator comprehension...
        return (
            # Yielding a 2-tuple of:
            #
            # * The name of this runner's method excluding runner prefix.
            # * Each "SimPipeRunner" instance defined on this class.
            (
                strs.remove_prefix(
                    text=runner_method_name,
                    prefix=cls._RUNNER_METHOD_NAME_PREFIX),
                runner
            )
            # For each such runner's method name and instance.
            for runner_method_name, runner in objects.iter_attrs_matching(
                obj=cls,
                predicate=lambda attr_name, attr_value: (
                    isinstance(attr_value, SimPipeRunner))))

    # ..................{ INITIALIZERS                       }..................
    @type_check
    def __init__(
        self,

        # Mandatory parameters.
        phase: SimPhase,

        #FIXME: For maintainability, refactor all of the following
        #"label_"-prefixed properties into concrete properties. The
        #"label_singular" property should default to returning simply "export";
        #the remaining properties should default to the expected strings.
        label_singular: str,

        # Optional parameters.
        label_plural: str = None,
        label_verb: str = 'Running',
    ) -> None:
        '''
        Initialize this pipeline.

        Parameters
        ----------
        phase : SimPhase
            Current simulation phase.
        label_singular : str
            Human-readable singular noun synopsizing the type of runners
            implemented by this subclass (e.g., ``animation``, ``plot``),
            ideally but *not* necessarily lowercase.
        label_plural : optional[str]
            Human-readable plural noun synopsizing the type of runners
            implemented by this subclass (e.g., ``animations``, ``plots``),
            ideally but *not* necessarily lowercase. Defaults to the passed
            ``label_singular`` parameter suffixed by an ``s`` character.
        label_verb: optional[str]
            Human-readable plural noun synopsizing the type of action performed
            by runners implemented by this subclass (e.g., ``Saving``), ideally
            but *not* necessarily capitalized. Defaults to ``Running``.
        '''

        # Default all unpassed parameters.
        if label_plural is None:
            label_plural = label_singular + 's'

        # Classify all passed parameters.
        self._phase = phase
        self._label_singular_lowercase = label_singular
        self._label_plural_lowercase = label_plural
        self._label_verb = label_verb

        # Human-readable capitalized singular noun.
        self._label_singular_uppercase = strs.uppercase_char_first(
            self._label_singular_lowercase)

    # ..................{ PROPERTIES                         }..................
    @property
    def is_enabled(self) -> bool:
        '''
        ``True`` only if the :meth:`run` method should run this pipeline.

        Specifically, if this boolean is:

        * ``False``, the :meth:`run` method reduces to a noop.
        * ``True``, this method behaves as expected (i.e., calls all currently
          enabled runner methods).

        Defaults to ``True``. Pipeline subclasses typically override this
        property to return a boolean derived from the simulation configuration
        file associated with the current phase.
        '''

        return True

    # ..................{ RUNNERS                            }..................
    def run(self) -> None:
        '''
        Run all currently enabled pipeline runners if this pipeline itself is
        currently enabled *or* noop otherwise.

        Specifically:

        * If the :meth:`is_enabled` property is ``True`` (implying this pipeline
          to be currently enabled):
          * For each :class:`YamlListItemTypedABC` instance (corresponding to
            the subconfiguration of a currently enabled pipeline runner) in the
            sequence of these instances listed by the :meth:`_runners_conf`
            property:
            * Call this method, passed this configuration.
            * If this method reports this runner's requirements to be
              unsatisfied (e.g., due to the current simulation configuration
              disabling extracellular spaces), this runner is ignored with a
              non-fatal warning.
        * Else, log an informative message and return immediately.
        '''

        # If this pipeline is disabled, log this fact and return immediately.
        if not self.is_enabled:
            logs.log_info('Excluding %s...', self._label_plural_lowercase)
            return
        # Else, this pipeline is enabled.

        # Log this pipeline run.
        logs.log_info(
            '%s %s...', self._label_verb, self._label_plural_lowercase)

        # For the object encapsulating all input arguments to be passed to each
        # currently enabled runner in this pipeline...
        for runner_conf in self._runners_conf:
            if not isinstance(runner_conf, YamlListItemTypedABC):
                raise BetseSimPipeException(
                    '_runners_conf() item {!r} '
                    'not instance of "YamlListItemTypedABC".'.format(
                        runner_conf))

            # If this runner is disabled, log this fact and ignore this runner.
            if not runner_conf.is_enabled:
                logs.log_debug(
                    'Ignoring disabled %s "%s"...',
                    self._label_singular_lowercase,
                    runner_conf.name)
                continue
            # Else, this runner is enabled.

            # Name of the method implementing this runner.
            runner_method_name = (
                self._RUNNER_METHOD_NAME_PREFIX + runner_conf.name)

            # Method implementing this runner *OR* None if this runner is
            # unrecognized.
            runner_method = objects.get_method_or_none(
                obj=self, method_name=runner_method_name)

            # If this runner is unrecognized, raise an exception.
            if runner_method is None:
                raise BetseSimPipeException(
                    '{} "{}" unrecognized.'.format(
                        self._label_singular_uppercase, runner_conf.name))
            # Else, this runner is recognized.

            # Attempt to pass this runner these arguments.
            try:
                runner_method(runner_conf)
            # If this runner's requirements are unsatisfied (e.g., due to the
            # current simulation configuration disabling extracellular spaces),
            # ignore this runner with a non-fatal warning and continue.
            except BetseSimPipeRunnerUnsatisfiedException as exception:
                logs.log_warning(
                    'Excluding %s "%s", as %s.',
                    self._label_singular_lowercase,
                    runner_conf.name,
                    exception.reason)

    # ..................{ EXCEPTIONS                         }..................
    @type_check
    def die_unless_runner_satisfied(
        self, runner: SimPipeRunner) -> None:
        '''
        Raise an exception if the passed runner is **unsatisfied** (i.e.,
        requires one or more simulation features disabled for the current
        simulation phase) *or* log an attempt to run this runner otherwise.

        Parameters
        ----------
        runner : SimPipeRunner
            Simulation pipeline runner to be tested.

        Raises
        ----------
        BetseSimVisualUnsatisfiedException
            If this runner is unsatisfied.
        '''

        # Strip the prefixing verb from the name of this runner's method (e.g.,
        # "export_" from "export_voltage_total"), raising a human-readable
        # exception if this name has no such prefix.
        runner_name = strs.remove_prefix(
            text=runner.method_name,
            prefix=self._RUNNER_METHOD_NAME_PREFIX,
            exception_message=(
                'Runner method name "{}" not prefixed by "{}".'.format(
                    runner.method_name, self._RUNNER_METHOD_NAME_PREFIX)))

        # If any runner requirement is unsatisfied, raise an exception.
        for requirement in runner.requirements:
            if not requirement.is_satisfied(phase=self._phase):
                raise BetseSimPipeRunnerUnsatisfiedException(
                    result='{} "{}" requirement unsatisfied'.format(
                        self._label_singular_uppercase, runner_name),
                    reason='{} disabled'.format(requirement.name),
                )
        # Else, all runner requirements are satisfied.

        # Log the subsequent attempt to run this runner.
        logs.log_info(
            '%s %s "%s"...',
            self._label_verb, self._label_singular_lowercase, runner_name)

    # ..................{ PRIVATE ~ subclass                 }..................
    @abstractproperty
    def _runners_conf(self) -> IterableTypes:
        '''
        Iterable of all :class:`YamlListItemTypedABC` instances for the current
        pipeline, each encapsulating all input parameters to be passed to the
        method implementing a runner currently contained in this pipeline.

        Pipeline subclasses typically implement this property to return an
        instance of the :class:``YamlList`` class listing all runners listed
        by the simulation configuration file associated with the current phase.
        '''

        pass

# ....................{ SUBCLASSES                         }....................
class SimPipeExportABC(SimPipeABC):
    '''
    Abstract base class of all **simulation export pipelines** (i.e., subclasses
    iteritavely exporting all variations on a single type of simulation export,
    either in parallel *or* in series).
    '''

    # ..................{ SUPERCLASS ~ constants             }..................
    _RUNNER_METHOD_NAME_PREFIX = 'export_'
    '''
    Substring prefixing the name of each runner defined by this pipeline.
    '''

    # ..................{ INITIALIZERS                       }..................
    def __init__(self, *args, **kwargs) -> None:

        # Initialize our superclass with all passed parameters and an
        # exportation-specific verb.
        super().__init__(*args, label_verb='Exporting', **kwargs)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level operating system (OS) facilities.

Caveats
----------
**Operating system-specific logic is poor form.** Do so *only* where necessary.
'''

# ....................{ IMPORTS                            }....................
import os, platform, sys
from betse.util.type.decorator.decmemo import func_cached
from betse.util.type.mapping.mapcls import OrderedArgsDict

# ....................{ TESTERS ~ posix                    }....................
@func_cached
def is_posix() -> bool:
    '''
    ``True`` only if the current operating system complies with POSIX standards
    (e.g., as required for POSIX-compliant symbolic link support).

    Typically, this implies this system to _not_ be vanilla Microsoft Windows
    and hence to be either:

    * A genuinely POSIX-compliant system.
    * A Cygwin-based Windows application (e.g., CLI terminal, GUI application).
    '''

    return os.name == 'posix'


@func_cached
def is_linux() -> bool:
    '''
    ``True`` only if the current operating system is Linux.
    '''

    return platform.system() == 'Linux'


@func_cached
def is_macos() -> bool:
    '''
    ``True`` only if the current operating system is Apple macOS, the operating
    system previously known as "macOS."
    '''

    return platform.system() == 'Darwin'

# ....................{ TESTERS ~ windows                  }....................
@func_cached
def is_windows() -> bool:
    '''
    ``True`` only if the current operating system is Microsoft Windows.

    This function reports `True` for both vanilla and Cygwin Microsoft Windows.
    '''
    return is_windows_vanilla() or is_windows_cygwin()


@func_cached
def is_windows_cygwin() -> bool:
    '''
    ``True`` only if the current operating system is **Cygwin Microsoft
    Windows** (i.e., running the Cygwin POSIX compatibility layer).
    '''
    return sys.platform == 'cygwin'


@func_cached
def is_windows_vanilla() -> bool:
    '''
    ``True`` only if the current operating system is **vanilla Microsoft
    Windows** (i.e., *not* running the Cygwin POSIX compatibility layer).
    '''
    return sys.platform == 'win32'

# ....................{ GETTERS                            }....................
@func_cached
def get_name() -> str:
    '''
    Human-readable name of the current operating system.

    This function returns:

    * Under Linux, the name of this Linux distribution, detected as follows:
      * If the `platform.linux_distribution()` function is available, the first
        element of the 3-tuple returned by this function (e.g., `CentOS`).
      * Else, the string returned by the `platform.system()` function. Since
        this is probably the generic string `Linux`, this is only a fallback.
    * Under macOS, `macOS`.
    * Under Cygwin Windows, `Windows (Cygwin)`. Note, however, that Cygwin
      masquerades as Linux, replicating the POSIX process model and GNU/Linux
      userland to a remarkable degree -- subject to the Big List of Dodgy Apps
      (BLODA), proprietary limitations of the Win32 API, and similar caveats.
    * Under non-Cygwin Windows, merely `Windows`.
    * Under all other platforms, the string returned by the `platform.system()`
      function.
    '''

    # Name to be returned, defaulting to that returned by platform.system().
    # Since this typically corresponds to the low-level name of the current
    # kernel (e.g., "Darwin", "Linux") rather than the high-level name of the
    # current OS (e.g., "macOS", "CentOS"), this is only a fallback.
    os_name = platform.system()

    # If the Linux-specific platform.linux_distribution() function is available,
    # return the first element of the 3-tuple "(distname, version, id)" returned
    # by this function (e.g., "CentOS"). Since platform.system() returns the
    # low-level kernel name "Linux", this name is ignored when feasible.
    if is_linux() and hasattr(platform, 'linux_distribution'):
        os_name = platform.linux_distribution()[0]
    # If macOS, return "macOS". Since platform.system() returns the low-level
    # kernel name "Darwin", this name is ignored.
    elif is_macos():
        os_name = 'macOS'
    # If Cygwin Windows, return "Windows (Cygwin)". Since platform.system()
    # returns a non-human-readable low-level uppercase label specific to the
    # word size of the current Python interpreter (e.g., "CYGWIN_NT-5.1*"), this
    # label is ignored.
    elif is_windows_cygwin():
        os_name = 'Windows (Cygwin)'
    # If non-Cygwin Windows, return merely "Windows". The name returned by
    # platform.system() depends on the current Windows version as follows:
    #
    # * If this is Windows Vista, this name is "Microsoft".
    # * Else, this name is "Windows".
    #
    # Hence, this name is ignored.
    elif is_windows_vanilla():
        os_name = 'Windows'

    # Return this name.
    return os_name


@func_cached
def get_version() -> str:
    '''
    Human-readable `.`-delimited version specifier string of the current
    operating system.

    This function returns:

    * Under Linux, the version reported by this Linux distribution as follows:
      * If the `platform.linux_distribution()` function is available, the second
        element of the 3-tuple returned by this function (e.g., `6.4`).
      * Else, the string returned by the `platform.release()` function. Since
        this is probably the non-human-readable `.`- and `-`-delimited version
        specifier string of the current Linux kernel (e.g., `4.1.15-gentoo-r1`),
        this is only a fallback.
    * Under macOS, this system's current major and minor version (e.g., `10.9`).
    * Under Windows, this system's current major version (e.g., `8`).
      Unfortunately, there appears to be no consistently reliable means of
      obtaining this system's current major _and_ minor version (e.g., `8.1`).
    * Under all other platforms, the string returned by the `platform.release()`
      function.
    '''

    # Version specifier to be returned, defaulting to that returned by
    # platform.release(). Since this typically corresponds to the low-level
    # version of the current kernel (e.g., "4.1.15") rather than the high-level
    # version of the current OS (e.g., "6.4"), this is only a fallback.
    os_version = platform.release()

    # If the Linux-specific platform.linux_distribution() function is available,
    # return the second element of the 3-tuple "(distname, version, id)"
    # returned by this function (e.g., "6.4"). Since platform.release() returns
    # the low-level kernel version (e.g., "4.1.15"), this version is ignored
    # when feasible.
    if is_linux() and hasattr(platform, 'linux_distribution'):
        os_version = platform.linux_distribution()[1]
    # If macOS *AND* the platform.mac_ver() function is available, return the
    # first element of the 3-tuple returned by this function (e.g., "10.9").
    # Since platform.release() returns the low-level kernel version
    # (e.g., "13.0.0"), this version is ignored when feasible.
    elif is_macos() and hasattr(platform, 'mac_ver'):
        # platform.mac_ver() returns a 3-tuple "(release, versioninfo, machine)"
        # of macOS-specific metadata, where "versioninfo" is itself a 3-tuple
        # "(version, dev_stage, non_release_version)". Return the high-level
        # "release" element (e.g., "10.9") rather than the optionally defined
        # low-level "version" element of the "versioninfo" element, which is
        # typically the empty string and hence useless.
        os_version = platform.mac_ver()[0]
    # If Windows, accept the default version specifier returned by
    # platform.release() (e.g., "8", "10"). Since this specifier appears to
    # coincide exactly with the first element of the 4-tuple returned by the
    # conditionally available platform.win32_ver() function, there is no benefit
    # to the latter approach.

    # Return this version.
    return os_version

# ....................{ GETTERS ~ metadata                 }....................
def get_metadata() -> OrderedArgsDict:
    '''
    Ordered dictionary synopsizing the current operating system.
    '''

    # Return this dictionary.
    return OrderedArgsDict(
        'name', get_name(),
        'version', get_version(),
    )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level support facilities for Yet Another Markup Language (YAML), the
file format encapsulating most input and output data for this application.
'''

#FIXME: Consider contributing various portions of this submodule back to
#"ruamel.yaml" -- particularly the Numpy-type-to-YAML-native-type conversions.

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To permit YAML implementations to be conditionally imported at
# application startup, no implementations (e.g., the top-level "yaml" package
# corresponding to PyYAML) are importable here.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

from betse import metadeps
from betse.exceptions import BetseYamlException
from betse.lib import libs
from betse.util.io import iofiles
from betse.util.io.log import logs
from betse.util.path import pathnames
from betse.util.type.obj import objects
from betse.util.type.types import (
    type_check, FileType, MappingOrSequenceTypes, ModuleType)
from enum import Enum

# ....................{ ENUMS                              }....................
YamlType = Enum('YamlType', ('PyYAML', 'ruamel_yaml',))
'''
Enumeration of all third-party YAML implementations supported by this submodule.
'''


YAML_TYPE_ACTIVE = None
'''
:data:`YamlType` member corresponding to the currently active third-party YAML
implementation if any *or* ``None`` if this submodule has yet to be initialized.
'''

# ....................{ GLOBALS                            }....................
FILETYPES = {'yaml', 'yml',}
'''
Set of all YAML-compliant filetypes.
'''

# ....................{ WARNERS                            }....................
@type_check
def warn_unless_filetype_yaml(filename: str) -> None:
    '''
    Log a non-fatal warning unless the passed filename has a YAML-compliant
    filetype (i.e., is suffixed by either ``.yaml`` or ``.yml``).

    Parameters
    ----------
    filename : str
        Absolute or relative path of this file.
    '''

    # Filetype of this file.
    filetype = pathnames.get_filetype_undotted_or_none(filename)

    # If this filetype is *NOT* YAML-compliant...
    if filetype not in FILETYPES:
        # If this file has *NO* filetype, log an appropriate warning.
        if filetype is None:
            logs.log_warning('YAML file "%s" has no filetype.', filename)
        # Else, this file has a filetype. Log an appropriate warning.
        else:
            logs.log_warning(
                'YAML file "%s" filetype "%s" neither "yaml" nor "yml".',
                filename, filetype)

# ....................{ LOADERS                            }....................
@type_check
def load(filename: str) -> MappingOrSequenceTypes:
    '''
    Load (i.e., open and read, deserialize) and return the contents of the
    YAML-formatted file with the passed path as either a dictionary or list
    via the active YAML implementation.

    Parameters
    ----------
    filename : str
        Absolute or relative path of this file.

    Returns
    ----------
    MappingOrSequenceTypes
        Dictionary or list corresponding to the contents of this file.
    '''

    # If this filename has no YAML-compliant filetype, log a warning.
    warn_unless_filetype_yaml(filename)

    # With this YAML file opened for character-oriented reading...
    with iofiles.reading_chars(filename) as yaml_file:
        # Load this YAML file via the active YAML implementation.
        if YAML_TYPE_ACTIVE is YamlType.PyYAML:
            return _load_pyyaml(yaml_file)
        elif YAML_TYPE_ACTIVE is YamlType.ruamel_yaml:
            return _load_ruamel(yaml_file)
        else:
            raise BetseYamlException(
                'YAML type "{}" unrecognized.'.format(YAML_TYPE_ACTIVE))


@type_check
def _load_pyyaml(yaml_file: FileType) -> MappingOrSequenceTypes:
    '''
    Load and return the contents of the YAML-formatted file with the passed
    readable file handle via the PyYAML parser (i.e., the :mod:`yaml` package).
    '''

    # PyYAML, validated and imported dynamically for safety.
    pyyaml = _import_pyyaml()

    # Load and return the contents of this YAML file.
    return pyyaml.load(yaml_file)


@type_check
def _load_ruamel(yaml_file: FileType) -> MappingOrSequenceTypes:
    '''
    Load and return the contents of the YAML-formatted file with the passed
    readable file handle via the :mod:`ruamel.yaml` parser.
    '''

    # Safe roundtripping YAML parser.
    ruamel_parser = _make_ruamel_parser()

    # Load and return the contents of this YAML file.
    return ruamel_parser.load(yaml_file)

# ....................{ SAVERS                             }....................
@type_check
def save(
    # Mandatory parameters.
    container: MappingOrSequenceTypes,
    filename: str,

    # Optional parameters.
    is_overwritable: bool = False,
) -> None:
    '''
    Save (i.e., open and write, serialize) the passed dictionary or list to the
    YAML-formatted file with the passed path via the active YAML implementation.

    Parameters
    ----------
    container: MappingOrSequenceTypes
        Dictionary or list to be written as the contents of this file.
    filename : str
        Absolute or relative path of this file.
    is_overwritable : optional[bool]
        ``True`` if overwriting this file when this file already exists *or*
        ``False`` if raising an exception when this file already exists.
        Defaults to ``False`` for safety.
    '''

    # If this filename has no YAML-compliant filetype, log a warning.
    warn_unless_filetype_yaml(filename)

    # With this YAML file opened for character-oriented writing...
    with iofiles.writing_chars(
        filename=filename, is_overwritable=is_overwritable) as yaml_file:
        # Save this YAML file via the active YAML implementation.
        if YAML_TYPE_ACTIVE is YamlType.PyYAML:
            return _save_pyyaml(container, yaml_file)
        elif YAML_TYPE_ACTIVE is YamlType.ruamel_yaml:
            return _save_ruamel(container, yaml_file)
        else:
            raise BetseYamlException(
                'YAML type "{}" unrecognized.'.format(YAML_TYPE_ACTIVE))

@type_check
def _save_pyyaml(
    container: MappingOrSequenceTypes, yaml_file: FileType) -> None:
    '''
    Save the passed container to the YAML-formatted file with the passed
    writable file handle via the PyYAML parser (i.e., the :mod:`yaml` package).
    '''

    # PyYAML, validated and imported dynamically for safety.
    pyyaml = _import_pyyaml()

    # Save this container to this YAML file.
    pyyaml.dump(
        data=container,
        stream=yaml_file,

        # Permit this parser to roundtrip Unicode characters.
        allow_unicode=True,
        encoding=None,

        # Coerce this parser into serializing nested collections as blocks
        # rather than inline "flow". See _make_ruamel_parser() for details.
        default_flow_style=False,
    )


@type_check
def _save_ruamel(
    container: MappingOrSequenceTypes, yaml_file: FileType) -> None:
    '''
    Save the passed container to the YAML-formatted file with the passed
    writable file handle via the :mod:`ruamel.yaml` parser.
    '''

    # Fully-qualified name of the module defining this container's subclass.
    container_class_module_name = objects.get_class_module_name(container)

    # If this container is *NOT* a "ruamel.yaml"-specific object returned by a
    # prior call to the _load_ruamel() function, log a non-fatal warning. While
    # this edge case does *NOT* constitute a fatal warning, it does disable
    # roundtripped preservation of comments and whitespace contained in the
    # original YAML file -- the principal motivation for preferring
    # "ruamel.yaml" over PyYAML.
    if not container_class_module_name.startswith('ruamel.'):
        logs.log_warning(
            'Non-"ruamel.yaml" type "%s.%s" not roundtrippable.',
            container_class_module_name, container.__class__.__name__)

    # Safe roundtripping YAML parser.
    ruamel_parser = _make_ruamel_parser()

    # Save this container to this YAML file.
    ruamel_parser.dump(container, yaml_file)

# ....................{ INITIALIZERS                       }....................
def init() -> None:
    '''
    Initialize both this submodule *and* the currently active YAML
    implementation (e.g., :mod:`ruamel.yaml`, PyYAML).

    This function selects the first such implementation importable under the
    active Python interpreter from the following list (in descending order of
    preference): :mod:`ruamel.yaml`, PyYAML.
    '''

    # Globals assigned to below.
    global YAML_TYPE_ACTIVE

    # Log this initialization.
    logs.log_debug(
        'Initializing preferred YAML binding "%s"...',
        metadeps.RUNTIME_MANDATORY_YAML_PROJECT_NAME)

    # Convert the non-typesafe setuptools-specific project name of this
    # implementation into a typesafe enumeration member.
    if metadeps.RUNTIME_MANDATORY_YAML_PROJECT_NAME == 'PyYAML':
        YAML_TYPE_ACTIVE = YamlType.PyYAML
    elif metadeps.RUNTIME_MANDATORY_YAML_PROJECT_NAME == 'ruamel.yaml':
        YAML_TYPE_ACTIVE = YamlType.ruamel_yaml
    else:
        raise BetseYamlException(
            'YAML project name "{}" unrecognized.'.format(
                metadeps.RUNTIME_MANDATORY_YAML_PROJECT_NAME))

    # If the active YAML implementation is PyYAML, initialize PyYAML.
    #
    # Since "ruamel.yaml" prefers a modern object-oriented API and hence is
    # locally initialized on object construction rather than globally on module
    # importation, note that "ruamel.yaml" requires no such initialization.
    if YAML_TYPE_ACTIVE is YamlType.PyYAML:
        _init_pyyaml()


def _init_pyyaml() -> None:
    '''
    Initialize PyYAML as the currently active YAML implementation.
    '''

    # Avoid circular import dependencies.
    from betse.lib.yaml import yamlrepr

    # PyYAML, validated and imported dynamically for safety.
    pyyaml = _import_pyyaml()

    # Globally add all representers required by this application to PyYAML.
    yamlrepr.add_representers(pyyaml)

    # Monkeypatch the following PyYAML method to eliminate Numpy-specific
    # warnings. Note that "ruamel.yaml" has since addressed all
    # Numpy-specific warnings and thus does *NOT* require this monkeypatch.
    pyyaml.representer.SafeRepresenter.ignore_aliases = _pyyaml_ignore_aliases

# ....................{ IMPORTERS                          }....................
def _import_pyyaml() -> ModuleType:
    '''
    Dynamically validate, import, and return the top-level module for PyYAML.
    '''

    return libs.import_runtime_optional('PyYAML')


def _import_ruamel() -> ModuleType:
    '''
    Dynamically validate, import, and return the top-level module for
    :mod:`ruamel.yaml`.
    '''

    return libs.import_runtime_optional('ruamel.yaml')

# ....................{ MAKERS                             }....................
def _make_ruamel_parser() -> 'ruamel.yaml.YAML':
    '''
    Safe roundtripping :mod:`ruamel.yaml` parser, where:

    * "Safe" implies this parser ignores all pragmas in this YAML file
      instructing parsers to construct arbitrary Python objects, whose YAML
      syntax is of the form: ``!!python/object:module.name { ... state ... }``.
    * "Roundtripping" implies the object deserialized from this YAML file
      preserves *all* comments and whitespace of this file, such that
      re-serializing this object back to disk preserves these substrings.
    '''

    # Avoid circular import dependencies.
    from betse.lib.yaml import yamlrepr

    # "ruamel.yaml", validated and imported dynamically for safety.
    ruamel_yaml = _import_ruamel()

    # Safe roundtripping YAML parser.
    ruamel_parser = ruamel_yaml.YAML(
        # Type of YAML parser to produce. Note that, by design, the
        # "rt" (i.e., roundtripping) parser is *ALWAYS* guaranteed to be safe.
        typ='rt',
    )

    # Permit this parser to roundtrip Unicode characters.
    ruamel_parser.allow_unicode = True

    # Coerce this parser into serializing nested collections as blocks rather
    # than inline "flow": e.g.,
    #
    #     # ...like this.
    #     b:
    #       c: 3
    #       d: 4
    #
    #     # ...not this.
    #     b: {c: 3, d: 4}
    ruamel_parser.default_flow_style = False

    # Locally add all representers required by this application to this parser.
    yamlrepr.add_representers(ruamel_parser.representer)

    # Return this parser.
    return ruamel_parser

# ....................{ MONKEYPATCHES                      }....................
def _pyyaml_ignore_aliases(self, data) -> bool:
    '''
    PyYAML-specific :meth:`yaml.representer.SafeRepresenter.ignore_aliases`
    method monkeypatched to eliminate Numpy-specific warnings.

    This method has been refactored to eliminate future warnings resembling:

        /usr/lib64/python3.4/site-packages/yaml/representer.py:135:
        FutureWarning: comparison to `None` will result in an elementwise object
        comparison in the future.
          if data in [None, ()]:
    '''

    # Refactor the default test of "if data in [None, ()]:" in the stock
    # ignore_aliases() method to avoid the future warning detailed above. To do
    # so safely, note that testing the passed "data" parameter against:
    #
    # * The singleton "None" object with the "is" operator is required.
    # * The non-singleton () object with the "==" rather than "is" operator is
    #   required. Bizarrely, despite the empty tuple being frozen, Python
    #   documentation notes that "...two occurrences of the empty tuple may or
    #   may not yield the same object". Hence, the comparison "data is ()" could
    #   unsafely return "False" even when "data" is the empty tuple!
    if data is None or data == ():
        return True
    if isinstance(data, (str, bytes, bool, int, float)):
        return True



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level operating system (OS)-specific display facilities.
'''

#FIXME: Submit a Stackoverflow answer encapsulating this logic. The
#osx.is_aqua() function in particular would be useful to a wide audience.

# ....................{ IMPORTS                            }....................
from betse.util.type.decorator.decmemo import func_cached
from betse.util.type.mapping.mapcls import OrderedArgsDict

# ....................{ TESTERS ~ head                     }....................
@func_cached
def is_headfull() -> bool:
    '''
    `True` only if the active Python interpreter is running **headfull** (i.e.,
    with access to a GUI display, the common case when running under a
    conventional desktop, laptop, or tablet device).
    '''

    # Avoid circular import dependencies.
    from betse.util.os import oses
    from betse.util.os.brand import macos
    from betse.util.os.shell import shellenv

    # The current process is headfull if and only if...
    return (
        # This is Windows, the current process is *ALMOST* certainly headfull.
        # While certain server-specific variants of Windows can and often are
        # run headless (e.g., Windows Nano Server), there appears to be no known
        # means of reliably distinguishing a headless from headfull Windows
        # environment in pure Python. For simplicity, the latter is assumed.
        oses.is_windows() or

        # Else, this is a POSIX-compatible platform.
        #
        # Since all POSIX-compatible platforms of interest support the headfull
        # X11 display server, we efficiently test for the accessibility of this
        # server via the ${DISPLAY} environment variable inherited from the
        # parent shell environment first.
        shellenv.is_var('DISPLAY') or

        # Else, all possible alternative display servers specific to the current
        # platform *MUST* be iteratively tested for.
        #
        # If this is Linux, the only remaining display servers are:
        #
        # * Mir, accessible via the ${MIR_SOCKET} environment variable.
        # * Wayland, accessible via the ${WAYLAND_DISPLAY} environment variable.
        #
        # Ergo, the current process is headfull if and only if one of these
        # variables is inherited from the parent shell environment.
        (oses.is_linux() and shellenv.is_var('MIR_SOCKET', 'WAYLAND_DISPLAY', )) or

        # If this is OS X, the only remaining display server is Aqua.
        (oses.is_macos() and macos.is_aqua())

        # Else, this platform is unrecognized. For safety, this platform is
        # assumed to be headless.
    )


def is_headless() -> bool:
    '''
    `True` only if the active Python interpreter is running **headless** (i.e.,
    with _no_ access to a GUI display, often due to running remotely over an
    SSH-encrypted connection supporting only CLI input and output).
    '''

    return not is_headfull()    # Makes sense.

# ....................{ TESTERS                            }....................
@func_cached
def is_linux_wayland() -> bool:
    '''
    `True` only if the active Python interpreter is running under a Wayland
    compositor-enabled Linux distribution.

    Caveats
    ----------
    For sanity, this function incorrectly assumes *all* Wayland compositors to
    comply with the X Desktop Group (XDG) standard by exporting the
    ``${XDG_SESSION_TYPE}`` environment variable with a value of ``wayland``. As
    this is *not* necessarily the case, this function may return false negatives
    for edge-case Wayland compositors.
    '''

    # Avoid circular import dependencies.
    from betse.util.os import oses
    from betse.util.os.shell import shellenv

    # If the current platform is *NOT* Linux, return False.
    if not oses.is_linux():
        return False
    # Else, the current platform is Linux.

    # String value of the ${XDG_SESSION_TYPE} environment variable if defined or
    # "None" otherwise.
    xdg_session_type = shellenv.get_var_or_none('XDG_SESSION_TYPE')

    # Return True only if this value is that of a Wayland compositor.
    return xdg_session_type == 'wayland'

# ....................{ GETTERS ~ metadata                 }....................
def get_metadata() -> OrderedArgsDict:
    '''
    Ordered dictionary synopsizing the current display.
    '''

    # Return this dictionary.
    return OrderedArgsDict(
        'headless', is_headless(),
        'wayland',  is_linux_wayland(),
    )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Unit tests for the :mod:`betse.util.path.dirs` submodule.
'''

# ....................{ IMPORTS                            }....................

# ....................{ TESTS                              }....................
def test_dirs_get_mtime_newest(betse_temp_dir: 'LocalPath') -> None:
    '''
    Unit test the :func:`betse.util.path.dirs.get_mtime_recursive_newest` and
    related :func:`betse.util.path.paths.get_mtime_recursive_newest` functions with a
    directory tree fabricated in the passed temporary directory.

    Parameters
    ----------
    betse_temp_dir : LocalPath
        Object encapsulating a temporary directory isolated to the current test.
    '''

    # Defer heavyweight imports.
    from betse.util.path import dirs, paths

    # Absolute paths of a subdirectory and file with arbitrary basenames
    # residing in this temporary directory.
    subdirpath  = betse_temp_dir.join("Quarks_Bar")
    subfilepath = betse_temp_dir.join("Grill")

    # Absolute paths of a subdirectory and two files with arbitrary basenames
    # residing in the above subdirectory residing in this temporary directory.
    subsubdirpath   = subdirpath.join("Gaming_House")
    subsubfilepath1 = subdirpath.join("And")
    subsubfilepath2 = subdirpath.join("Holosuite_Arcade")

    # Create these subdirectories and files, ensuring (...get it, "ensuring"?)
    # that the last such path is that asserted to be the most recent below
    subdirpath.ensure(dir=True)
    subfilepath.ensure(file=True)
    subsubdirpath.ensure(dir=True)
    subsubfilepath1.ensure(file=True)
    subsubfilepath2.ensure(file=True)

    # Relevant paths as strings rather than "LocalPath" objects.
    dirname = str(betse_temp_dir)
    subsubdirname = str(subsubdirpath)
    subsubfilename2 = str(subsubfilepath2)

    # Ensure the last path created above to be the most recent.
    assert (
        dirs.get_mtime_recursive_newest(dirname) ==
        paths.get_mtime_nonrecursive(subsubfilename2)
    )

    # Update the mtime of an arbitrary subdirectory.
    subsubdirpath.setmtime()

    # Ensure this subdirectory to now be the most recent.
    assert (
        dirs.get_mtime_recursive_newest(dirname) ==
        paths.get_mtime_nonrecursive(subsubdirname)
    )

    # Ensure this subdirectory to now be the most recent when queried through
    # the related paths.get_mtime_recursive_newest() function.
    assert (
        paths.get_mtime_recursive_newest(
            (dirname, subsubdirname, subsubfilename2)) ==
        paths.get_mtime_nonrecursive(subsubdirname)
    )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
**Root test configuration** (i.e., early-time configuration guaranteed to be run
by :mod:`pytest` *before* passed command-line arguments are parsed) for this
test suite.

Caveats
----------
For safety, this configuration should contain *only* early-time hooks absolutely
required by :mod:`pytest` design to be defined in this configuration. Hooks for
which this is the case (e.g., :func:`pytest_addoption`) are explicitly annotated
as such in official :mod:`pytest` documentation with a note resembling:

    Note

    This function should be implemented only in plugins or ``conftest.py`` files
    situated at the tests root directory due to how pytest discovers plugins
    during startup.

This file is the aforementioned ``conftest.py`` file "...situated at the tests
root directory."
'''

# ....................{ IMPORTS                            }....................

# ....................{ HOOKS ~ option                     }....................
def pytest_addoption(parser: '_pytest.config.Parser') -> None:
    '''
    Hook run immediately on :mod:`pytest` startup *before* parsing command-line
    arguments, typically registering test suite-specific :mod:`argparse`-style
    options and ini-style config values.

    Options
    ----------
    After :mod:`pytest` parses these options, the globally accessible
    :attr:`pytest.config.option.{option_var_name}` attribute provides the value
    of the argument accepted by each option (if any), where
    ``{option_var_name}`` is the value of the ``dest`` keyword argument passed
    to the :meth:`parser.add_option` method in the body of this hook.

    Specifically, the following option variables are guaranteed to be defined:

    * :attr:`pytest.config.option.export_sim_conf_dirname`, the value of the
      ``--export-sim-conf-dir`` command-line option if passed *or* ``None``
      otherwise.

    Caveats
    ----------
    This hook should be implemented *only* in plugins or ``conftest.py`` files
    situated at the top-level tests directory for this application (e.g., like
    the current file), due to plugin discovery by :mod:`pytest` at startup.

    Parameters
    ----------
    parser : _pytest.config.Parser
        :mod:`pytest`-specific command-line argument parser, inspired by the
        :mod:`argparse` API.
    '''

    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # CAUTION: The name of this option is assumed to *NOT* change across Git
    # commits.  Changing this name would violate this assumption and hence
    # forwards compatibility with future versions of this test suite.
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    # String argument options (i.e., options requiring a string argument),
    # disabled unless explicitly passed.
    parser.addoption(
        '--export-sim-conf-dir',
        dest='export_sim_conf_dirname',
        default=None,
        help=(
            'target directory into which all '
            'source simulation configuration directories produced by '
            '"@skip_unless_export_sim_conf"-marked tests are to be copied'
        ),
        metavar='DIRNAME',
    )

# ....................{ HOOKS ~ session                    }....................
#FIXME: This hook doesn't actually appear to be invoked. Deprecated, perhaps?
def pytest_sessionstart(session):
    '''
    Hook run immediately *before* starting the current test session (i.e.,
    calling the :func:`pytest.session.main` function).
    '''

    pass


#FIXME: This hook doesn't actually appear to be invoked. Deprecated, perhaps?
def pytest_sessionfinish(session, exitstatus):
    '''
    Hook run immediately *after* completing the current test session (i.e.,
    calling the :func:`pytest.session.main` function).
    '''

    pass



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level **Python-based multithreading** (i.e., multithreading implemented at
the Python API level rather than by third-party C extensions, libraries, or
packages) facilities.
'''

# ....................{ IMPORTS                            }....................
from betse.exceptions import BetsePyException

# ....................{ EXCEPTIONS                         }....................
def die_unless_threadable() -> None:
    '''
    Raise an exception unless the active Python interpreter supports
    multithreading.

    Raises
    ----------
    BetsePyException
        If this interpreter does *not* support multithreading.

    See Also
    ----------
    :func:`is_threadable`
        Further details.
    '''

    # If this Python interpreter is *NOT* multithreadable, raise an exception.
    if not is_threadable():
        raise BetsePyException(
            'Python interpreter not multithreadable '
            '(i.e., "_thread" module not importable).')

# ....................{ TESTERS                            }....................
def is_threadable() -> bool:
    '''
    ``True`` only if the active Python interpreter supports multithreading.

    Equivalently, this function returns ``True`` only if the low-level standard
    :mod:`_thread` module is importable under this interpreter. This module is
    strictly optional and hence available only for systems both supporting
    multithreading *and* compiling Python with such support.

    Systems supporting multithreading includes:

    * Windows.
    * macOS.
    * Linux.
    * SGI IRIX.
    * Solaris 2.x.
    * Systems implementing the POSIX-compliant ``pthread`` API.
    '''

    # Avoid circular import dependencies.
    from betse.util.type import modules

    # Return true only if a low-level optional module installed *ONLY* when
    # Python supports multithreading is importable.
    return modules.is_module('_thread')



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Fixtures creating temporary directories isolated for safety to specific tests.
'''

# ....................{ IMPORTS                            }....................
from betse_test.util import requests
from pytest import fixture

# ....................{ FIXTURES                           }....................
# Test-scope fixture creating and returning a new object for each discrete test.
@fixture
def betse_temp_dir(
    request: '_pytest.python.FixtureRequest',
    tmpdir_factory: '_pytest.tmpdir.tmpdir_factory',
) -> 'LocalPath':
    '''
    Per-test fixture creating a temporary directory isolated to the test
    requesting this fixture and returning an object encapsulating this
    directory.

    This directory is guaranteed to be specific to this test. The basename of
    this directory is the name of this test excluding the prefixing substring
    ``test_``. When requested by the ``test_cli_sim_default`` test, for example,
    this fixture creates a temporary directory ``{tmpdir}/cli_sim_default`` for
    the absolute path ``{tmpdir}`` of this test session's root temporary
    directory (e.g., ``/tmp/pytest-0/cli_sim_default``).

    This directory is safely accessible *only* for the duration of this test.
    Subsequently run tests and fixtures *cannot* safely reuse this directory,
    although doing so is technically feasible in unreliable edge-cases.

    Parameters
    ----------
    request : _pytest.python.FixtureRequest
        Builtin fixture describing the parent fixture or test of this fixture.
    tmpdir_factory : _pytest.tmpdir.tmpdir_factory
        Builtin session-scoped fixture whose ``mktemp()`` method returns a
        :class:`py.path.local` instance encapsulating a new temporary directory.

    Returns
    ----------
    LocalPath
        Object encapsulating this temporary directory.
    '''

    # Defer heavyweight imports.
    from betse.util.type.text import strs

    # Name of the current test.
    test_name = requests.get_tested_name(request)
    # print('    request.node: {}'.format(request.node))
    # print('    test_name: {}'.format(test_name))

    # Basename of the temporary directory containing this configuration file,
    # set to the name of the current test excluding the prefixing "test_".
    temp_dir_basename = strs.remove_prefix(
        text=test_name,
        prefix='test_',
        exception_message=(
            'Test name "{}" not prefixed by "test_".'.format(test_name)),
    )

    # Create this temporary directory and wrap this directory's absolute path
    # with a high-level "py.path.local" object. See also:
    #     http://pytest.org/latest/tmpdir.html#the-tmpdir-factory-fixture
    temp_dirpath = tmpdir_factory.mktemp(temp_dir_basename)

    # Return this object.
    return temp_dirpath



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level standard output facilities.
'''

#FIXME: Copy setup.util.get_command_output() here after completion.

# ....................{ IMPORTS                            }....................

# ....................{ OUTPUTTERS                         }....................
def output(*objects) -> None:
    '''
    Print all passed objects to stdout _without_ logging these objects.

    This function is provided only for orthogonality with the function of the
    same name defined by the `stderr` module.

    This function is intentionally _not_ named `print()`. Doing so introduces
    subtle issues elsewhere.
    '''

    print(*objects)


def output_lines(*objects) -> None:
    '''
    Print all passed objects to standard output *without* logging such objects,
    delimiting each such object by a newline.
    '''

    print('\n'.join(*objects))


def output_traceback() -> None:
    '''
    Print the current call stack to standard output.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.call import callers

    # Print this call stack, excluding the calls to both this and the
    # callers.get_traceback() functions.
    output(callers.get_traceback(-2))



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Main entry point of this application's command line interface (CLI).

This submodule is a thin wrapper intended to be:

* Indirectly imported and run from external entry point scripts installed by
  setuptools (e.g., the ``betse`` command).
* Directly imported and run from the command line (e.g., via
  ``python -m betse.cli``).
'''

# ....................{ IMPORTS                            }....................
from betse.cli.climain import BetseCLI
from betse.util.path.command import cmdexit

# ....................{ MAIN                               }....................
def main(arg_list: list = None) -> int:
    '''
    Run this application's command-line interface (CLI) with the passed
    arguments if non-``None`` *or* with the arguments passed on the command line
    (i.e., :attr:`sys.argv`) otherwise.

    This function is provided as a convenience to callers requiring procedural
    functions rather than conventional methods (e.g., :mod:`setuptools`).

    Parameters
    ----------
    arg_list : list
        List of zero or more arguments to pass to this interface. Defaults to
        ``None``, in which case arguments passed on the command line (i.e.,
        :attr:`sys.argv`) will be used instead.

    Returns
    ----------
    int
        Exit status of this interface and hence this process as an unsigned byte
        (i.e., integer in the range ``[0, 255]``).
    '''

    return BetseCLI().run(arg_list)

# ....................{ MAIN                               }....................
# If this module is imported from the command line, run this application's CLI;
# else, noop. For POSIX compliance, the exit status returned by this function
# is propagated to the caller as this script's exit status.
if __name__ == '__main__':
    cmdexit.exit(main())



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **callback** (i.e., caller-defined object external to this
application whose methods are called by other methods internal to this
application as a means of facilitating inter-package communication)
functionality.
'''

# ....................{ IMPORTS                            }....................
from abc import ABCMeta  #, abstractmethod
from betse.exceptions import BetseCallbackException
# from betse.util.io.log import logs
from betse.util.type.types import type_check  #, NoneType

# ....................{ SUPERCLASSES                       }....................
class CallbacksABC(metaclass=ABCMeta):
    '''
    Abstract base class of all **callbacks** (i.e., caller-defined object
    external to this application whose methods are called by other methods
    internal to this application as a means of facilitating inter-package
    communication) subclasses.

    Design
    ----------
    Each of the public methods defined by this superclass is a callback
    intended to be called *only* by the **source callable** (i.e., function or
    method calling each such callback). The source callback receives this
    object from the **sink callable** (i.e., function or method instantiating
    this object and passing this object to the source callable).

    Attributes
    ----------
    _progress_min : int
        Minimum value subsequently passed to the :meth:`progressed` callback by
        the source callable. While typically 0, the sink callable should *not*
        necessarily assume this to be the case.
    _progress_max : int
        Maximum value subsequently passed to the :meth:`progressed` callback by
        the source callable.
    _progress_next : int
        Value to be subsequently passed to the :meth:`progressed` callback by
        the :meth:`progressed_next` callback.
    '''

    # ..................{ INITIALIZERS                       }..................
    def __init__(self) -> None:
        '''
        Initialize this callbacks object.
        '''

        # Nullify all instance variables for safety.
        self._clear_progress()

    # ..................{ CALLBACKS ~ progress               }..................
    @type_check
    def progress_ranged(
        self, progress_max: int, progress_min: int = 0) -> None:
        '''
        Callback passed the range of progress values subsequently passed to the
        :meth:`progressed` callback by the source callable.

        Design
        ----------
        The source callable should call this callback exactly once per call
        *before* either performing any significant work or calling the
        :meth:`progressed` callback. Doing so enables the sink callable to
        initialize external objects (e.g., progress bars) requiring this range.

        Subclasses are required to reimplement this callback and call this
        superclass implementation in their own reimplementation.

        Parameters
        ----------
        progress_min : optional[int]
            Minimum value subsequently passed to the :meth:`progressed`
            callback by the source callable. While typically 0, the sink
            callable should *not* necessarily assume this to be the case. To
            guarantee that the ``progress_max`` parameter is exactly equal to
            the number of times that the :meth:`progressed_next` callback is
            called by the source callable, this parameter should ideally be 0.
            Defaults to 0.
        progress_max : int
            Maximum value subsequently passed to the :meth:`progressed`
            callback by the source callable. If the ``progress_min`` parameter
            is 0, this is equivalent to the number of times that the
            :meth:`progressed_next` callback is called by the source callable.

        Raises
        ----------
        :class:`BetseCallbackException`
            If either:

            * The source callable is still performing work (i.e., the
              :meth:`progressed` callback has yet to be called with a progress
              value equal to ``progress_max``).
            * ``progress_min`` is greater than or equal to ``progress_max``.
        '''

        # If this range is invalid, raise an exception.
        if progress_min > progress_max:
            raise BetseCallbackException(
                'Minimum progress {} exceeds maximum progress {}.'.format(
                    progress_min, progress_max))
        elif progress_min == progress_max:
            raise BetseCallbackException(
                'Minimum progress {} equals maximum progress {}.'.format(
                    progress_min, progress_max))

        # If the source callable is still performing work, raise an exception.
        self._die_if_progressing()

        # Classify all passed parameters for subsequent reference by the
        # progressed() callback.
        self._progress_min = progress_min
        self._progress_max = progress_max

        # Value to be subsequently passed to the progressed() callback by the
        # first call to the progressed_next() callback. See the
        # progressed_next() callback for a justification of the "+ 1" here.
        self._progress_next = progress_min + 1
        # logs.log_debug('next progress: %d', self._progress_next)


    @type_check
    def progressed(self, progress: int) -> None:
        '''
        Callback passed the current state of progress for work performed by the
        source callable.

        If this is the last possible progress value (i.e., if the passed
        ``progress`` parameter is equal to the ``progress_max`` parameter
        previously passed to the :meth:`progress_ranged` callback), this method
        additionally records the source callable to no longer be performing
        work by calling the :meth:`_clear_progress` method. This preserves
        contractual guarantees, including the guarantee that the next call to
        either this or the :meth:`progressed_next` callback will raise an
        exception.

        Design
        ----------
        The source callable should call this callback repeatedly while
        performing any significant work. Doing so enables the sink callable to
        incrementally update external objects (e.g., progress bars).

        Subclasses are required to reimplement this callback and call this
        superclass implementation in their own reimplementation.

        Parameters
        ----------
        progress : int
            Current state of progress for work performed by the source
            callable. If this integer is *not* in the range previously passed
            to the :meth:`progress_ranged` callback, an exception is raised. If
            this is the first call to this callback by the source callable,
            this integer should ideally be one larger than the minimum progress
            value previously passed to the :meth:`progress_ranged` callback.
            See the :meth:`progressed_next` callback for further discussion.

        Raises
        ----------
        :class:`BetseCallbackException`
            If either:

            * The :meth:`progress_ranged` callback has yet to be called.
            * This progress is *not* in the range previously passed to the
              :meth:`progress_ranged` callback by this source callable.

        See Also
        ----------
        :meth:`progressed_next`
            Higher-level callback automatically passing the next progress value
            according to the the range previously passed to the
            :meth:`progress_ranged` callback.
        '''

        # If progress_ranged() has yet to be called, raise an exception.
        self._die_unless_progressing()

        # If this progress is *NOT* in the range previously passed to
        # progress_ranged(), raise an exception.
        if not (self._progress_min <= progress <= self._progress_max):
            raise BetseCallbackException(
                'Progress {} not in range [{}, {}].'.format(
                    progress, self._progress_min, self._progress_max))

        # If this is the last possible progress value, record the source
        # callable to no longer be performing work.
        if progress == self._progress_max:
            self._clear_progress()


    def progressed_next(self) -> None:
        '''
        Higher-level callback wrapping the lower-level :meth:`progressed`
        callback by automatically passing the next progress value to that
        callback according to the the range previously passed to the
        :meth:`progress_ranged` callback.

        Guarantees
        ----------
        The first call to this callback is guaranteed to emit one greater than
        the minimum progress value previously passed to the
        :meth:`progress_ranged` callback. While arguably unintuitive, this
        behaviour corresponds with expectations in the sink callable.

        Assuming that the minimum progress value previously passed to the
        :meth:`progress_ranged` callback was 0 as is the common case, this
        behaviour has the additional advantage of guaranteeing that the maximum
        progress value previously passed to that callback is exactly equal to
        the number of times that this :meth:`progressed_next` callback is
        called by the source callable.

        Specifically, since the sink callable typically handles the
        :meth:`progress_ranged` callback by initializing a progress bar to the
        passed range, the sink callable is assumed to have already "consumed"
        the first progress value for all intents and purposes; instructing the
        sink callable to "reconsume" the same progress value would effectively
        hide the actual first iteration of work performed by the source
        callable from the end user perspective.

        Design
        ----------
        For generality, the sink callable should typically *only* handle the
        :meth:`progressed` callback. This callback is merely syntactic sugar
        conveniently simplifying the implementation of the source callback
        calling this callback.

        Raises
        ----------
        :class:`BetseCallbackException`
            If the :meth:`progress_ranged` callback has yet to be called.
        '''

        # If progress_ranged() has yet to be called, raise an exception.
        self._die_unless_progressing()

        # Notify the sink callback of the current state of progress.
        self.progressed(progress=self._progress_next)

        # If the source callback has work remaining to perform, increment this
        # state *AFTER* notifying this callback.
        if self._progress_next is not None:
            self._progress_next += 1

    # ..................{ EXCEPTIONS                         }..................
    def _die_if_progressing(self) -> None:
        '''
        Raise an exception if the source callable is still performing work.

        Raises
        ----------
        :class:`BetseCallbackException`
            If the source callable is currently performing work.

        See Also
        ----------
        :meth:`_is_progressing`
            Further details.
        '''

        if self._is_progressing:
            raise BetseCallbackException(
                'progress_ranged() callback called before '
                'progressed() callback called with maximum progress value.')


    def _die_unless_progressing(self) -> None:
        '''
        Raise an exception unless the source callable is still performing work.

        Raises
        ----------
        :class:`BetseCallbackException`
            If the source callable is *not* currently performing work.

        See Also
        ----------
        :meth:`_is_progressing`
            Further details.
        '''

        if not self._is_progressing:
            raise BetseCallbackException(
                'progressed() callback called before '
                'progress_ranged() callback.')

    # ..................{ TESTERS                            }..................
    @property
    def _is_progressing(self) -> bool:
        '''
        ``True`` only if the source callable is still performing work.

        Equivalently, this method returns ``False`` only:

        * If the source callable is *not* currently performing work.
        * If the state of work performed by this callable has been cleared
          (i.e., to ``None`` values) rather than set to non-``None`` integers.
        * Unless the :meth:`progress_ranged` callback has been called more
          recently than the :meth:`_clear_progress` method.
        '''

        return (
            self._progress_min  is not None and
            self._progress_max  is not None and
            self._progress_next is not None
        )

    # ..................{ CLEARERS                           }..................
    def _clear_progress(self) -> None:
        '''
        Record the source callable to no longer be performing work by clearing
        (i.e., resetting) the state of work performed by this callable to
        ``None`` values.
        '''

        # Nullify all instance variables for safety.
        self._progress_min = None
        self._progress_max = None
        self._progress_next = None



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level facilities for serializing and deserializing Numpy arrays to and from
comma-separated value (CSV) files.
'''

# ....................{ IMPORTS                            }....................
import numpy as np
from betse.exceptions import BetseSequenceException, BetseStrException
from betse.util.io.log import logs
from betse.util.path import dirs
from betse.util.type import types
from betse.util.type.text import strs
from betse.util.type.types import type_check
from collections import OrderedDict

# ....................{ WRITERS                            }....................
#FIXME: Donate this function back to Numpy as a new np.savecsv() function
#paralleling the existing np.savetxt() function.
@type_check
def write_csv(filename: str, column_name_to_values: OrderedDict) -> None:
    '''
    Serialize each key-value pair of the passed ordered dictionary into a new
    column in comma-separated value (CSV) format to the plaintext file with the
    passed filename.

    Caveats
    ----------
    To ensure that all columns of this file have the same number of rows, all
    values of this dictionary *must* be one-dimensional sequences of:

    * The same length. If this is not the case, an exception is raised.
    * Any type satisfying the :class:`SequenceTypes` API, including:
      * Numpy arrays.
      * Lists.
      * Tuples.

    Typically, each such value is a one-dimensional Numpy array of floats.

    Parameters
    ----------
    filename : str
        Absolute or relative path of the plaintext file to be written. If this
        file already exists, this file is silently overwritten.
    column_name_to_values: OrderedDict
        Ordered dictionary of all columns to be serialized such that:
        * Each key of this dictionary is a **column name** (i.e., terse string
          describing the type of data contained in this column).
        * Each value of this dictionary is **column data** (i.e.,
          one-dimensional sequence of all arbitrary data comprising this
          column).

    Raises
    ----------
    BetseSequenceException
        If one or more values of this dictionary are either:
        * *Not* sequences.
        * Sequences whose length differs from that of any preceding value
          sequences of this dictionary.
    BetseStrException
        If this column name contains one or more characters reserved for use by
        the CSV non-standard, including:
        * Double quotes, reserved for use as the CSV quoting character.
        * Newlines, reserved for use as the CSV row delimiting character.
    '''

    # Log this serialization.
    logs.log_debug('Writing CSV file: %s', filename)

    # Validate the contents of this dictionary. While the np.column_stack()
    # function called below also does, the exceptions raised by the latter are
    # both ambiguous and non-human-readable and hence effectively useless.
    #
    # Length of all prior columns or None if no columns have yet to be iterated.
    columns_prior_len = None

    # List of all column names sanitized such that each name containing one or
    # more comma characters is double-quoted.
    column_names = []

    # For each passed column...
    for column_name, column_values in column_name_to_values.items():
        # If this column is *NOT* a sequence, raise a human-readable exception.
        if not types.is_sequence_nonstr(column_values):
            raise BetseSequenceException(
                'Column "{}" type {!r} not a sequence.'.format(
                    column_name, type(column_values)))

        # Length of this column.
        column_len = len(column_values)

        # If this is the first column to be iterated, require all subsequent
        # columns be of the same length.
        if columns_prior_len is None:
            columns_prior_len = column_len
        # Else if this column's length differs from that of all prior columns,
        # raise a human-readable exception.
        elif column_len != columns_prior_len:
            raise BetseSequenceException(
                'Column "{}" length {} differs from '
                'length {} of prior columns.'.format(
                    column_name, column_len, columns_prior_len))

        # If this column name contains one or more reserved characters, raise an
        # exception. This includes:
        #
        # * Double quotes, reserved for use as the CSV quoting character.
        # * Newlines, reserved for use as the CSV row delimiting character.
        if '"' in column_name:
            raise BetseStrException(
                'Column name {} contains '
                "one or more reserved '\"' characters.".format(column_name))
        if '\n' in column_name:
            raise BetseStrException(
                'Column name {} contains '
                'one or more newline characters.'.format(column_name))

        # If this column name contains one or more commas (reserved for use as
        # the CSV delimiter), double-quote this name. Since the prior logic
        # guarantees this name to *NOT* contain double quotes, no further logic
        # is required
        if ',' in column_name:
            column_name = '"{}"'.format(column_name)

        # Append this sanitized column name to this list of such names.
        column_names.append(column_name)

    # Comma-separated string listing all column names.
    columns_name = strs.join_on(column_names, delimiter=',')

    # Two-dimensional Numpy array of all row data converted from this column
    # data, whose:
    #
    # * First dimension indexes each sampled time step such that each element is
    #   a one-dimensional Numpy array of length the number of columns (i.e., the
    #   number of key-value pairs in the passed dictionary).
    # * Second dimension indexes each column data point for this time step
    columns_values = np.column_stack(column_name_to_values.values())

    # Create the directory containing this file if needed.
    dirs.make_parent_unless_dir(filename)

    # Serialize these sequences to this file in CSV format.
    np.savetxt(
        fname=filename,
        X=columns_values,
        header=columns_name,
        delimiter=',',

        # Prevent Numpy from prefixing the above header by "# ". Most popular
        # software importing CSV files implicitly supports a comma-delimited
        # first line listing all column names.
        comments='',
    )
